/**
 * BeanPot Bean Container Project.
 * Copyright (C) 2009  Rubecula Software, LLC.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Module: BeanConfiguration.java
 * Created on Apr 6, 2010
 * @version $Revision: 1.1 $
 */

package com.rubecula.beanpot;

import java.io.File;
import java.net.URL;

import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.XMLConfiguration;

/**
 * This class extends the XMLConfiguration class of the Jakarta Commons
 * Configuration package. Here are the "rules" for the validation by DTD of
 * configuration files:
 * <ul>
 * <li>by default, no validation of configuration files is performed;</li>
 * <li>validation of configuration files can be forced in one of two ways:</li>
 * <ul>
 * <li>a DOCTYPE entity can be defined in the configuration file;</li>
 * <li>the method {@link BeanPot#setValidate(boolean)} can be called with true.
 * In this case the URL used for the DTD is defined by the
 * {@link BeanPot#setDTD(java.net.URL)} method which defaults to
 * "../../..//beanpot.dtd" which corresponds to
 * <i>main/resources/beanpot.dtd</i>.</li>
 * </ul>
 * </ul>
 * <p>
 * Note that the normal way of changing the validation parameters is to call
 * {@link BeanPot#setValidate(boolean, URL)} or one of its shortcut signatures.
 * This in turn calls {@link BeanPot#setDTD(java.net.URL)}.
 * </p>
 * 
 * @author Robin Hillyard
 * 
 */
public class BeanConfiguration extends XMLConfiguration {

	/**
	 * @return a newly constructed XMLConfiguration
	 */
	public static BeanConfiguration create() {
		final BeanConfiguration result = new BeanConfiguration();
		result.registerEntityId(ENTITY_KEY, BeanPot.getDTD());
		result.setValidating(BeanPot.isValidate());
		return result;
	}

	/**
	 * @param configFile
	 * @return a newly created {@link XMLConfiguration} object, loaded from the
	 *         given configFile.
	 * @throws ConfigurationException
	 */
	public static BeanConfiguration create(final File configFile) throws ConfigurationException {
		final BeanConfiguration configuration = create();
		configuration.setFile(configFile);
		configuration.load();
		return configuration;
	}

	/**
	 * @param filename
	 * @return a newly created {@link XMLConfiguration} with filename loaded.
	 * @throws ConfigurationException
	 */
	public static XMLConfiguration create(final String filename) throws ConfigurationException {
		final XMLConfiguration configuration = create();
		configuration.setFileName(filename);
		configuration.load();
		return configuration;
	}

	/**
	 * @param url
	 * @return a new XMLConfiguration
	 * @throws ConfigurationException
	 */
	public static BeanConfiguration create(final URL url) throws ConfigurationException {
		final BeanConfiguration configuration = create();
		configuration.setURL(url);
		configuration.load();
		return configuration;
	}

	/**
	 * 
	 */
	BeanConfiguration() {
		super();
	}

	/**
	 * 
	 */
	private static final long serialVersionUID = -3640459021206730922L;

	/**
	 * This is the key that is used for retrieving the appropriate DTD.
	 */
	public static final String ENTITY_KEY = "BeanPot"; //$NON-NLS-1$
}
