/**
 * BeanPot Bean Container Project.
 * Copyright (C) 2009  Rubecula Software, LLC.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Module: BeanMap.java
 * Created on Mar 11, 2009
 * @version $Revision: 1.1 $
 */

package com.rubecula.beanpot;

import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.HierarchicalConfiguration;
import org.apache.commons.configuration.SubnodeConfiguration;
import org.apache.commons.configuration.XMLConfiguration;
import org.apache.commons.logging.Log;

/**
 * This type represents the bean map used by the {@link BeanPot} container. It
 * also contains all of the utilities for dealing with beans themselves. Since
 * the beans are, more or less, by definition, extensions only of Object, much
 * of the work on beans cannot be extracted into a specific type. Instead, the
 * utilities are actually private methods, some of which are class methods.
 * 
 * BeanMap is an extension of {@link HashMap}<String,Object>.
 * 
 * For more information about configuring properties, constructor args, etc.
 * please see the javadoc for the following methods:
 * <ul>
 * <li>
 * {@link #configureBean(HierarchicalConfiguration)}</li>
 * <li>{@link #getPropertyValue(String, Class, Configuration)}</li>
 * <li>{@link #addBeanRefToArgs(List, SubnodeConfiguration)}</li>
 * </ul>
 * <br/>
 * XXX currently we share the logger with the BeanPot class.
 * 
 * @author Robin Hillyard
 * 
 */
public class BeanMap extends HashMap<String, Object> {

	/**
	 * @return the beanpropertypattern
	 */
	public static Pattern getBeanPropertyPattern() {
		return BEANPROPERTYPATTERN;
	}

	/**
	 * @param requiredClass
	 *            the class that we require
	 * @param string
	 *            the String that we hope represents an object of class
	 *            requiredClass
	 * @return the object of class requiredClass which corresponds to the string
	 *         by virtue of invoking the class method valueOf(String)
	 * @throws BeanPotException
	 * 
	 *             TODO consider moving this into BeanUtilities
	 */
	static Object getValueFromString(final Class<?> requiredClass, final String string) throws BeanPotException {
		if (requiredClass.equals(String.class))
			return string;
		try {
			final Method valueOf = requiredClass.getMethod(M_VALUE_OF, String.class);
			return valueOf.invoke(NULL_OBJECT, string);
		} catch (final NoSuchMethodException e) {
			throw new BeanPotException(
					"getValueFromString(): list/map value " + string + ": is not of the specified type (" + requiredClass + ") and cannot be converted by a valueOf(String) method."); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		} catch (final IllegalAccessException e) {
			throw new BeanPotException(
					"getValueFromString(): list/map value " + string + ": access to valueOf(String) method is illegal", e); //$NON-NLS-1$ //$NON-NLS-2$
		} catch (final InvocationTargetException e) {
			throw new BeanPotException(
					"getValueFromString(): list/map value " + string + ": exception thrown by valueOf(String) method.", e.getTargetException()); //$NON-NLS-1$ //$NON-NLS-2$
		}
	}

	/**
	 * Construct a new BeanMap with default values for initialCapacity and
	 * loadFactor
	 */
	public BeanMap() {
		super();
		initialize();
	}

	/**
	 * Construct a new BeanMap with default value for loadFactor and given value
	 * for initialCapacity.
	 * 
	 * @param initialCapacity
	 */
	public BeanMap(final int initialCapacity) {
		super(initialCapacity);
		initialize();
	}

	/**
	 * Construct a new BeanMap with given values for initialCapacity and
	 * loadFactor
	 * 
	 * @param initialCapacity
	 * @param loadFactor
	 */
	public BeanMap(final int initialCapacity, final float loadFactor) {
		super(initialCapacity, loadFactor);
		initialize();
	}

	/**
	 * Add a bean to this bean map.
	 * 
	 * XXX Add side-effect: if the bean implements the Nameable interface, that
	 * the key is used to actually name the bean.
	 * 
	 * @param key
	 *            the key by which this bean will be identified (usually its
	 *            identifier).
	 * @param bean
	 *            the bean itself
	 * @param registerAsMBean
	 *            if true we try to register as an MBean
	 * @return the previous version of bean with this key (else null).
	 * @throws BeanPotException
	 * @see Map#put(Object, Object)
	 */
	public Object addBean(final String key, final Object bean, final Kleenean registerAsMBean) throws BeanPotException {
		if (registerAsMBean != Kleenean.False) {
			if (BeanPot.getMBeanPot() != null) {
				try {
					BeanPot.getMBeanPot().registerMBean(key, bean);
				} catch (final BeanPotException e) {
					if (registerAsMBean == Kleenean.True)
						throw e;
					getLogger().debug("Bean for key " + key + " is not an MBean"); //$NON-NLS-1$ //$NON-NLS-2$
				}
			} else if (registerAsMBean == Kleenean.True)
				throw new BeanPotException("addBean(): Cannot register as MBean because no MBeanPot class available"); //$NON-NLS-1$
		}

		return put(key, bean);
	}

	/**
	 * Get the bean from this bean map, given the <code>key</code> (its
	 * identifier). If the bean is not found, a warning will be sent to the
	 * writer.
	 * 
	 * @param key
	 *            the identifier with which this bean was added to this bean
	 *            map.
	 * @return the bean else null (if no such bean).
	 * @see Map#get(Object)
	 */
	public Object getBean(final String key) {
		return getBean(key, true);
	}

	/**
	 * Get the bean from this bean map, given the <code>key</code> (its
	 * identifier). If the bean is not found, a warning will be sent to the
	 * writer.
	 * 
	 * @param key
	 *            the identifier with which this bean was added to this bean
	 *            map.
	 * @param warn
	 *            if true and the bean is non-existent then we will squawk
	 * @return the bean else null (if no such bean).
	 * @see Map#get(Object)
	 */
	public Object getBean(final String key, final boolean warn) {
		final Object bean = get(key);
		if (warn && bean == null)
			getLogger().warn("No such bean: " + key); //$NON-NLS-1$
		return bean;
	}

	/**
	 * Get the set of bean keys for this bean map.
	 * 
	 * @return a Set of String objects.
	 * @see Map#keySet()
	 */
	public Set<String> getBeanKeys() {
		return keySet();
	}

	/**
	 * @return the classLoader
	 */
	public ClassLoader getClassLoader() {
		return this.classLoader;
	}

	/**
	 * Method to get the beanPotLogger for this bean map.
	 * 
	 * @return the current beanPotLogger for this BeanMap
	 */
	public Log getLogger() {
		return this.logger;
	}

	/**
	 * Method to allow applications to dynamically update the value of a bean's
	 * property.
	 * 
	 * @param identifier
	 *            the identifier of the bean to update.
	 * @param name
	 *            the name of the property to update.
	 * @param value
	 *            the value the property should take on.
	 * @return the bean that has been updated.
	 * @throws BeanPotException
	 */
	public Object setBeanProperty(final String identifier, final String name, final Object value) throws BeanPotException {
		final Object bean = getBean(identifier);
		if (bean != null) {
			BeanUtilities.setBeanProperty(bean, value, name, identifier);
			return bean;
		}
		throw new BeanPotException("Bean not found: " + identifier); //$NON-NLS-1$
	}

	/**
	 * @param classLoader
	 *            the classLoader to set
	 */
	public void setClassLoader(final ClassLoader classLoader) {
		this.classLoader = classLoader;
	}

	/**
	 * Set the Bean Pot Logger for this BeanMap
	 * 
	 * @param logger
	 *            the new Bean Pot Logger
	 */
	public void setLogger(final Log logger) {
		getLogger().debug("Setting BeanMap logger to: " + logger); //$NON-NLS-1$
		this.logger = logger;
	}

	/**
	 * Given the appropriate <code>config</code> for a bean, add the bean to
	 * this bean map.
	 * 
	 * Constructor arguments are defined by tags of either this form:
	 * 
	 * <pre>
	 * &lt;constructor-arg&gt;
	 *     &lt;ref bean=&quot;Applet&quot; /&gt;
	 * &lt;/constructor-arg&gt;
	 * </pre>
	 * <p>
	 * Or of this form:
	 * </p>
	 * 
	 * <pre>
	 * &lt;constructor-arg class=&quot;java.lang.Double&quot;&gt;
	 *     &lt;value&gt;3.1415927&lt;/value&gt;
	 * &lt;/constructor-arg&gt;
	 * </pre>
	 * <p>
	 * and where the class is optional if the value is either an Integer or a
	 * String.
	 * 
	 * 
	 * @param config
	 *            a hierarchical configuration, typically from getting a
	 *            configuration "at" a key from a configuration, typically
	 *            instantiated as {@link XMLConfiguration}
	 * @throws BeanPotException
	 * 
	 * @see BeanUtilities#addBean(String, String, Object[], Kleenean, boolean,
	 *      Class)
	 */
	void addBean(final HierarchicalConfiguration config) throws BeanPotException {
		final String nodeName = config.getRootNode().getName();
		if (nodeName.equalsIgnoreCase(BeanUtilities.S_BEAN)) {
			addBean(config, nodeName); // void the result
		} else
			throw new BeanPotException("configuration node is not a bean: " + nodeName); //$NON-NLS-1$
	}

	/**
	 * Method to get a reference to an existing bean (from the subnode
	 * configuration <code>beanRef</code>) and add the bean to the set of
	 * constructor <code>args</code>.
	 * 
	 * @param args
	 *            a List of objects which will be passed to a constructor.
	 * @param beanRef
	 *            the configuration describing a reference to a bean.
	 * @throws BeanPotException
	 * @see #getBean(String, boolean)
	 */
	@SuppressWarnings("unchecked")
	void addBeanRefToArgs(final List<Object> args, final SubnodeConfiguration beanRef) throws BeanPotException {
		final Iterator<String> keys = beanRef.getKeys();
		while (keys.hasNext()) {
			final String key = keys.next();
			if (key.equalsIgnoreCase(S_ATTR_BEAN)) {
				final String refBean = beanRef.getString(key);
				final Object bean = refBean.equalsIgnoreCase(BeanPot.AGENT_WRITER_) ? BeanPot.getWriter() : getBean(refBean);
				if (bean != null)
					args.add(bean);
				else
					throw new BeanPotException("getConstructorArgs()" + //$NON-NLS-1$
							BeanUtilities.S_COLON + "referenced bean has not yet been defined: " + refBean); //$NON-NLS-1$
			}
		}
	}

	/**
	 * For each bean in the bean map, remove it.
	 * 
	 * @return
	 */
	boolean cleanup() {
		boolean ok = true;
		final Iterator<String> iterator = getBeanKeys().iterator();
		while (iterator.hasNext()) {
			final String key = iterator.next();
			ok &= closeBean(key);
			unregisterMBean(key);
			iterator.remove();
		}
		clear(); // should be unnecessary
		return ok;
	}

	private boolean closeBean(final String key) {
		final Object bean = getBean(key);
		if (bean instanceof Closeable)
			try {
				((Closeable) bean).close();
				return true;
			} catch (final BeanPotException e) {
				this.logger.error("error closing bean " + key, e); //$NON-NLS-1$
				return false;
			}
		return true;
	}

	/**
	 * Given the appropriate <code>config</code> for a bean, configure the
	 * (previously instantiated) bean in this bean map.
	 * 
	 * A property may be of one of the following types:
	 * <ul>
	 * <li><b>BigDecimal</b></li>
	 * <li><b>BigInteger</b></li>
	 * <li><b>Boolean</b></li>
	 * <li><b>Byte</b></li>
	 * <li><b>Double</b></li>
	 * <li><b>Float</b></li>
	 * <li><b>Integer</b></li>
	 * <li><b>List</b></li> (results in a Collection<?> object - see
	 * {@link BeanUtilities#getListParameter(String, Configuration)} ) for more
	 * detail
	 * <li><b>Long</b></li>
	 * <li><b>Map</b></li> (results in a Map<?,?> object - see
	 * {@link BeanUtilities#getMapParameter(HierarchicalConfiguration)} ) for
	 * more detail
	 * <li><b>Object</b> (in the case of a referenced bean)</li>
	 * <li><b>Short</b></li>
	 * <li><b>String</b></li>
	 * </ul>
	 * <p>
	 * A configurable bean property is defined as one where there is a matching
	 * pair of get/set methods (although the set method may not be visible) and
	 * where the type of the return/parameter is derivable using methods such as
	 * {@link Configuration#getBoolean(String)} from the string value.
	 * 
	 * These properties are expressed in the configuration file by tags of the
	 * form:
	 * </p>
	 * 
	 * <pre>
	 * &lt;property name=&quot;color&quot;&gt;
	 *     &lt;value&gt;Red&lt;/value&gt;
	 * &lt;/property&gt;
	 * </pre>
	 * <p>
	 * Or, in the case of a list property (where a list property extends
	 * {@link Collection}&lt;?&gt; where the Generic type ? can be explicitly
	 * specified by using the <i>valueclass</i> attribute), as follows:
	 * </p>
	 * 
	 * <pre>
	 * &lt;property name=&quot;colors&quot; valueclass=&quot;String&quot;&gt;
	 *     &lt;list&gt;
	 *         &lt;value&gt;Red&lt;/value&gt;
	 *         &lt;value&gt;White&lt;/value&gt;
	 *         &lt;value&gt;Blue&lt;/value&gt;
	 *     &lt;/list&gt;
	 * &lt;/property&gt;
	 * </pre>
	 * <p>
	 * This method can yield a collection of objects of any type which provides
	 * a public class method with name and signature:
	 * <code>valueOf(String)</code>, including but not limited to Integer, Long,
	 * Double, String. Additionally, a String which begins with a single "!"
	 * represents a bean defined elsewhere in the configuration file.
	 * </p>
	 * <p>
	 * Note particularly:
	 * <ul>
	 * <li>a string which should literally begin with "!" must be represented in
	 * the configuration file by substituting two "!" characters for the desired
	 * "!";</li>
	 * <li>since list elements by default are separated by ",", it is necessary
	 * that Strings which include a literal "," should be represented some other
	 * way -- there is no escape mechanism for the literal "," in a string.</li>
	 * </ul>
	 * </p>
	 * <p>
	 * Or, in the case of a map property (where a map property extends
	 * {@link Map}&lt;?,?&gt where the Generic types ? can be explicitly
	 * specified by using the <i>keyclass</i> and <i>valueclass</i> attributes),
	 * as follows:
	 * </p>
	 * 
	 * <pre>
	 * &lt;property name=&quot;colors&quot; keyclass=&quot;String&quot; valueclass=&quot;Integer&quot;&gt;
	 *     &lt;map&gt;
	 *         &lt;entry key=&quot;True&quot;&gt;1&quot;&lt;/entry&gt;
	 *         &lt;entry key=&quot;False&quot; value=&quot;0&quot; /&gt;
	 *     &lt;/map&gt;
	 * &lt;/property&gt;
	 * </pre>
	 * <p>
	 * Note that keyclass defaults to String. Note also that the keyclass and
	 * valueclass definitions can occur in the map entity instead of the
	 * property entity (map entity takes precedence if both define a class).
	 * Note also that there are two valid forms for a map entry, as shown in the
	 * example above.
	 * </p>
	 * <p>
	 * The following methods/fields are used:
	 * {@link BeanUtilities#configureBean(String, HierarchicalConfiguration)},
	 * {@link BeanUtilities#S_BEAN} (the configuration element to be matched),
	 * {@link BeanUtilities#S_ATTR_ID} (the attribute to define the identifier)
	 * </p>
	 * 
	 * @param config
	 *            a hierarchical configuration, typically from getting a
	 *            configuration "at" a key from a configuration, typically
	 *            instantiated as {@link XMLConfiguration}
	 * @throws BeanPotException
	 *             (various situations)
	 * 
	 */
	@SuppressWarnings("unchecked")
	void configureBean(final HierarchicalConfiguration config) throws BeanPotException {
		final String nodeName = config.getRootNode().getName();
		if (nodeName.equalsIgnoreCase(BeanUtilities.S_BEAN)) {
			final String id = getBeanId(config);
			if (id != null)
				try {
					// First configure the properties of the bean
					getBeanUtilities().configureBean(id, config);

					// Now add this bean as a member of other beans as
					// appropriate.
					for (final SubnodeConfiguration subnode : (List<SubnodeConfiguration>) config.configurationsAt(S_MEMBER)) {
						BeanUtilities.makeBeanAMember(this, subnode, id);
					}
				} catch (final IntrospectionException e) {
					throw new BeanPotException(M_CONFIGURE_BEANS + BeanUtilities.S_COLON + "introspection exception for bean" + //$NON-NLS-1$
							BeanUtilities.S_COLON + id, e);
				}
			else
				throw new BeanPotException(M_CONFIGURE_BEANS + BeanUtilities.S_COLON + "no id defined for node" + //$NON-NLS-1$
						BeanUtilities.S_COLON + nodeName);
		} else
			throw new BeanPotException(M_CONFIGURE_BEANS + BeanUtilities.S_COLON + "configuration node is not a bean" + //$NON-NLS-1$
					BeanUtilities.S_COLON + nodeName);
	}

	/**
	 * This private method configures the properties of the bean map. Note that
	 * the properties of the BeanPot itself cannot be set (the beanPot itself is
	 * not really a bean). Such properties are defined for example as follows:
	 * 
	 * <pre>
	 * &lt;properties&gt;
	 *     &lt;property name=&quot;logger&quot;&gt;
	 *         &lt;value&gt;!BeanPotLogger&lt;/value&gt;
	 *     &lt;/property&gt;
	 * &lt;/properties&gt;
	 * </pre>
	 * 
	 * @param hierarchicalConfiguration
	 *            the configuration
	 * 
	 * @throws BeanPotException
	 */
	@SuppressWarnings("unchecked")
	void configureBeanMapProperties(final HierarchicalConfiguration hierarchicalConfiguration) throws BeanPotException {
		final List<HierarchicalConfiguration> propertyConfigurations = hierarchicalConfiguration
				.configurationsAt(BeanMap.S_PROPERTIES);
		try {
			final PropertyDescriptor[] propertyDescriptors = Introspector.getBeanInfo(getClass()).getPropertyDescriptors();
			for (final HierarchicalConfiguration propertyConfig : propertyConfigurations) {
				// XXX fix this rather ugly construction
				getBeanUtilities().configureBeanProperties(this, propertyConfig, propertyDescriptors, "_beanMap_"); //$NON-NLS-1$
			}
		} catch (final IntrospectionException e) {
			throw new BeanPotException("BeanPot.configureBeanMapProperties(): introspection exception"); //$NON-NLS-1$
		}
	}

	/**
	 * Add the bean into the {@link #_externalBeans} map.
	 * 
	 * @param key
	 * @param bean
	 * @throws BeanPotException
	 *             if the key is already in use.
	 */
	void doImposeBean(final String key, final Object bean) throws BeanPotException {
		if (this._externalBeans.containsKey(key))
			throw new BeanPotException("external bean " + key + " already exists"); //$NON-NLS-1$ //$NON-NLS-2$
		this._externalBeans.put(key, bean);
	}

	/**
	 * If the string passed in starts with "!", then return the bean referenced
	 * (or a String). See {@link BeanUtilities#getStringAsBean(String)}
	 * 
	 * @param string
	 * @return either a bean or a String (or null)
	 */
	Object getBeanRef(final String string) {
		if (string.startsWith(BeanUtilities.S_SHRIEK))
			return getBeanUtilities().getStringAsBean(string);
		return null;
	}

	/**
	 * Get the value (as an Object of class <code>clazz</code>) corresponding to
	 * the property identified by <code>key</code> from the <code>config</code>.
	 * In the case that the value defined by <code>key</code> is a String, and
	 * <code>clazz</code> is not a String, then if the string starts with "!",
	 * it will be interpreted as a reference to a bean. Otherwise, an exception
	 * will be thrown.
	 * 
	 * @param key
	 *            the key defining the value in the context of the
	 *            <code>config</code>
	 * @param clazz
	 *            the class required for the property
	 * @param config
	 *            the configuration describing the property.
	 * @return the value as an Object
	 * @throws BeanPotException
	 *             the property is a String not beginning with "!" and the clazz
	 *             is not one of the named types above (in the return)
	 */
	@SuppressWarnings({ "boxing" })
	Object getPropertyValue(final String key, final Class<?> clazz, final Configuration config) throws BeanPotException {
		Object result = null;
		if (clazz == short.class || Short.class.isAssignableFrom(clazz)) {
			result = config.getShort(key);
		} else if (clazz == byte.class || Byte.class.isAssignableFrom(clazz)) {
			result = config.getByte(key);
		} else if (clazz == int.class || Integer.class.isAssignableFrom(clazz)) {
			result = config.getInt(key);
		} else if (clazz == long.class || Long.class.isAssignableFrom(clazz)) {
			result = config.getLong(key);
		} else if (clazz == double.class || Double.class.isAssignableFrom(clazz)) {
			result = config.getDouble(key);
		} else if (clazz == String.class) {
			result = config.getString(key);
		} else if (clazz == boolean.class || Boolean.class.isAssignableFrom(clazz)) {
			result = config.getBoolean(key);
		} else if (clazz == float.class || Float.class.isAssignableFrom(clazz)) {
			result = config.getFloat(key);
		} else if (BigInteger.class.isAssignableFrom(clazz)) {
			result = config.getBigInteger(key);
		} else if (BigDecimal.class.isAssignableFrom(clazz)) {
			result = config.getBigDecimal(key);
		} else if (Collection.class.isAssignableFrom(clazz)) {
			result = getBeanUtilities().getListParameter(BeanUtilities.S_LIST + key, config);
		} else if (clazz.isArray()) {
			result = getBeanUtilities().getArrayParameter(key, clazz, config);
		} else if (Map.class.isAssignableFrom(clazz)) {
			// XXX check this is required
			if (config instanceof HierarchicalConfiguration) {
				result = getBeanUtilities().getMapParameter((HierarchicalConfiguration) config);
			} else
				throw new BeanPotException(M_GET_PROPERTY_VALUE + BeanUtilities.S_COLON
						+ "configuration for Map is not hierarchical"); //$NON-NLS-1$
		} else {
			final String beanRef = config.getString(key);
			if (beanRef != null) {
				// TODO move the getBeanRef functionality into
				// getValueFromString
				result = getBeanRef(beanRef);
				if (result == null) {
					final Object valueFromString = getValueFromString(clazz, beanRef);
					if (valueFromString != null)
						return valueFromString;
					throw new BeanPotException(
							M_GET_PROPERTY_VALUE
									+ BeanUtilities.S_COLON
									+ "unsupported property type: " + clazz + "[note that references to beans in this context must be preceded by a '!'"); //$NON-NLS-1$ //$NON-NLS-2$
				}
			}
		}

		return result;
	}

	/**
	 * First call {@link MBeanPot#unregisterMBean(String)}. Then call
	 * {@link #remove(Object)}.
	 * 
	 * @param key
	 * @return the bean that has been removed.
	 */
	Object removeBean(final String key) {
		unregisterMBean(key);
		return remove(key);
	}

	/**
	 * For each bean in the bean map which implements the {@link Runnable}
	 * interface, create a new thread based on it, start the thread and then
	 * wait until that thread has finished.
	 */
	void runBeans() {
		for (final String key : getBeanKeys()) {
			final Object bean = BeanPot.getBean(key);
			if (bean instanceof Runnable) {
				BeanPot.getWriter().println("run bean: " + key); //$NON-NLS-1$
				final Thread t = new Thread((Runnable) bean);
				t.start();
				try {
					t.join();
					BeanPot.getWriter().println("bean finished: " + key); //$NON-NLS-1$
				} catch (final InterruptedException e) {
					BeanPot.getBeanPotLogger().debug("thread interrupt: " + e.getLocalizedMessage()); //$NON-NLS-1$					
				}
			}
		}
	}

	/**
	 * Decode the type, class, id and mbean designation from the
	 * <code>config</code> and call
	 * {@link #addBean(HierarchicalConfiguration, String, String, String, Kleenean, Type, boolean, Class)}
	 * .
	 * 
	 * @param config
	 * @param nodeName
	 * @return the bean
	 * @throws BeanPotException
	 */
	@SuppressWarnings("unchecked")
	private String addBean(final HierarchicalConfiguration config, final String nodeName) throws BeanPotException {
		final Iterator<String> keys = config.getKeys();
		String clazz = null;
		String id = null;
		Kleenean mBean = Kleenean.Maybe;
		Type type = Type.INSTANCE;
		boolean required = true;
		Class<? extends Throwable> acceptableException = null;
		while (keys.hasNext()) {
			final String key = keys.next();
			if (key.equalsIgnoreCase(BeanUtilities.S_ATTR_CLASS))
				clazz = config.getString(key);
			else if (key.equalsIgnoreCase(BeanUtilities.S_ATTR_ID))
				id = config.getString(key);
			else if (key.equalsIgnoreCase(S_ATTR_MBEAN))
				mBean = Kleenean.valueOf(config.getBoolean(key));
			else if (key.equalsIgnoreCase(BeanUtilities.S_ATTR_TYPE))
				type = Type.valueOf(config.getString(key).toUpperCase());
			else if (key.equalsIgnoreCase(BeanUtilities.S_ATTR_REQUIRED))
				required = config.getBoolean(key);
			else if (key.equalsIgnoreCase(BeanUtilities.S_ATTR_ACCEPTABLE_EXCEPTION) && config.getString(key).length() > 0)
				acceptableException = (Class<? extends Throwable>) BeanUtilities.getClassForString(config.getString(key),
						getClassLoader());
		}
		final Object bean = addBean(config, nodeName, clazz, id, mBean, type, required, acceptableException);
		return bean != null ? id : null;
	}

	/**
	 * According to type, add the bean in one of the following ways: external
	 * (by invoking {@link #addExternalBean(String, String, Kleenean)});
	 * singleton (by invoking
	 * {@link BeanUtilities#addBean(String, String, Object[], Kleenean, boolean, Class)}
	 * with constructor args null; instance (by invoking
	 * {@link BeanUtilities#getConstructorArgs(HierarchicalConfiguration)} and
	 * passing those to
	 * {@link BeanUtilities#addBean(String, String, Object[], Kleenean, boolean, Class)}
	 * .
	 * 
	 * @param config
	 * @param nodeName
	 * @param clazz
	 * @param key
	 *            the key by which the new bean will be referenced, i.e. the id
	 *            of the bean.
	 * @param mBean
	 * @param type
	 * @param required
	 *            XXX
	 * @param acceptableException
	 *            XXX
	 * @return the bean
	 * @throws BeanPotException
	 */
	private Object addBean(final HierarchicalConfiguration config, final String nodeName, final String clazz, final String key,
			final Kleenean mBean, final Type type, final boolean required, final Class<? extends Throwable> acceptableException)
			throws BeanPotException {
		if (key != null) {
			if (type == Type.EXTERNAL)
				return addExternalBean(key, clazz, mBean);

			try {
				final Object[] constructorArgs = type == Type.SINGLETON ? null : getBeanUtilities().getConstructorArgs(config);
				return getBeanUtilities().addBean(key, clazz, constructorArgs, mBean, required, acceptableException);
			} catch (final Exception e) {
				throw new BeanPotException("thrown adding bean: " + key, e); //$NON-NLS-1$
			}
		}
		throw new BeanPotException("no id defined for node: " + nodeName); //$NON-NLS-1$
	}

	/**
	 * @param key
	 *            the key
	 * @param clazz
	 *            the required class
	 * @param mBean
	 *            whether or not to register as an MBean
	 * @return
	 * @throws BeanPotException
	 */
	private Object addExternalBean(final String key, final String clazz, final Kleenean mBean) throws BeanPotException {
		final Object bean = getExternalBean(key);
		if (bean != null) {
			// TODO consider using checkCompatibleTypes()
			if (((Class<?>) BeanUtilities.getClassForString(clazz, getClassLoader())).isAssignableFrom(bean.getClass())) {
				addBean(key, bean, mBean); // Void the result
				return bean;
			}
			throw new BeanPotException("external bean " + key + " does not extend class: " + clazz); //$NON-NLS-1$//$NON-NLS-2$
		}
		throw new BeanPotException("external bean " + key + " not defined"); //$NON-NLS-1$//$NON-NLS-2$
	}

	/**
	 * @param config
	 * @return
	 */
	@SuppressWarnings("unchecked")
	private String getBeanId(final HierarchicalConfiguration config) {
		final Iterator<String> keys = config.getKeys();
		String id = null;
		while (keys.hasNext()) {
			final String key = keys.next();
			if (key.equalsIgnoreCase(BeanUtilities.S_ATTR_ID))
				id = config.getString(key);
		}
		return id;
	}

	/**
	 * @return the beanUtilities
	 */
	private BeanUtilities getBeanUtilities() {
		return this._beanUtilities;
	}

	/**
	 * Method to get the external bean referenced by key for the purpose of
	 * adding it to the bean map. This operation can only be performed once - as
	 * a side effect, after we retrieve the bean, we remove it from the
	 * {@link #_externalBeans} map, thus preventing multiple references within
	 * the bean map.
	 * 
	 * @param key
	 * @return the previously imposed external bean for the given key
	 * @throws BeanPotException
	 * @see java.util.Map#get(java.lang.Object)
	 */
	private Object getExternalBean(final Object key) throws BeanPotException {
		final Object result = this._externalBeans.get(key);
		this._externalBeans.remove(key);
		if (result != null)
			return result;
		throw new BeanPotException(
				"external bean " + key + " has not yet been externally imposed (or has already been defined in the configuration file)."); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * 
	 */
	private void initialize() {
		this._beanUtilities = new BeanUtilities(this);
	}

	/**
	 * @param key
	 */
	private void unregisterMBean(final String key) {
		if (BeanPot.getMBeanPot() != null)
			BeanPot.getMBeanPot().unregisterMBean(key);
	}

	/**
	 * This is used when invoking a class method.
	 */
	private static final Object NULL_OBJECT = null;

	static final private Pattern BEANPROPERTYPATTERN = Pattern.compile("([^\\.]+)\\.(\\p{Alpha}+)"); //$NON-NLS-1$

	/**
	 * member
	 */
	private static final String S_MEMBER = "member"; //$NON-NLS-1$

	/**
	 * valueOf
	 */
	private static final String M_VALUE_OF = "valueOf"; //$NON-NLS-1$

	/**
	 * getPropertyValue()
	 */
	private static final String M_GET_PROPERTY_VALUE = "getPropertyValue()"; //$NON-NLS-1$

	/**
	 * properties
	 */
	private static final String S_PROPERTIES = "properties"; //$NON-NLS-1$

	/**
	 * Attribute for specifying a management bean.
	 * 
	 * Attribute: mbean
	 */
	private static final String S_ATTR_MBEAN = "[@mbean]"; //$NON-NLS-1$

	/**
	 * Attribute for defining a referenced bean in a constructor argument.
	 * 
	 * Attribute: bean
	 */
	private static final String S_ATTR_BEAN = "[@bean]"; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final long serialVersionUID = -3810312961338396254L;

	/**
	 * configureBeans()
	 */
	private static final String M_CONFIGURE_BEANS = "configureBeans()"; //$NON-NLS-1$

	/**
	 * reference to the bean utilities class (for convenience)
	 */
	transient BeanUtilities _beanUtilities;

	/**
	 * An instance field to store the logger for this BeanMap. By default, the
	 * logger is the (static) logger for BeanPot.
	 */
	private transient Log logger = BeanPot.getBeanPotLogger();

	/**
	 * A map of externally imposed beans which is only useful before
	 * configuration. Externally imposed beans are added to this map via the
	 * method {@link #doImposeBean(String, Object)} and then referenced in here
	 * just like any other beans.
	 */
	private final Map<String, Object> _externalBeans = new HashMap<String, Object>();

	private transient ClassLoader classLoader = getClass().getClassLoader();

}
