/**
 * BeanPot Bean Container Project.
 * Copyright (C) 2009  Rubecula Software, LLC.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Module: BeanPot.java
 * Created on Mar 9, 2009
 * @version $Revision: 1.1 $
 */

package com.rubecula.beanpot;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Serializable;
import java.lang.management.ManagementFactory;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.regex.Matcher;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationBuilder;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.HierarchicalConfiguration;
import org.apache.commons.configuration.SubnodeConfiguration;
import org.apache.commons.configuration.XMLConfiguration;
import org.apache.commons.configuration.beanutils.BeanFactory;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.rubecula.beanpot.serialize.SerializerFactory;

/**
 * <p>
 * Singleton container class which configures, manages and runs beans.
 * Dependency injection is supported.
 * </p>
 * <p>
 * Beans are defined in an XML file (or some other file which can produce a
 * {@link HierarchicalConfiguration} object). Bean <i>properties</i> are defined
 * as properties. Bean properties may be references to other beans (the name
 * preceded by a "!").
 * </p>
 * <p>
 * Constructor arguments may be defined where the value is a reference to
 * another bean or where the value is an Object. Where the value is an object
 * whose value can be parsed from a String (according to the rules of the
 * Jakarta Commons Configuration package) the standard form of constructor arg
 * is used:
 * </p>
 * 
 * <pre>
 * &lt;constructor-arg class=&quot;Long&quot;&gt; &lt;value&gt;1&lt;/value&gt; &lt;/constructor-arg&gt;
 * </pre>
 * 
 * Note that the class attribute is optional if the value is an Integer or a
 * String. </p>
 * <p>
 * Where the reference is to another bean (which must have been defined
 * previously in the XML file), a special form of constructor arg is used:
 * </p>
 * 
 * <pre>
 * &lt;constructor-arg&gt; &lt;ref bean=&quot;My Bean&quot;/&gt; &lt;/constructor-arg&gt;
 * </pre>
 * <p>
 * Constructor arguments which are Lists or arrays are supported in much the
 * same was as list properties. However, you can only have one argument in the
 * constructor if it is going to be a list or an array. Be sure to specify the
 * class (such as java.util.List or Integer[]) and valueClass appropriately.
 * Note that at run-time there is no type-checking for a List (generics do not
 * apply at runtime) but there is type-checking for an array (so make sure that
 * all of the elements can be created with the correct type). See
 * testBeans13.xml for examples.
 * </p>
 * <p>
 * Constructor arguments which are Maps are not supported as <i>constructor
 * args</i>, although they are supported as properties.
 * </p>
 * <p>
 * Beans which implement the {@link Configurable} interface, will have their
 * {@link Configurable#preConfigure(Object)} and
 * {@link Configurable#postConfigure(Object)} methods invoked at the appropriate
 * time (before and after configuration).
 * </p>
 * <p>
 * Currently, all beans undergo the normal configuration (and dependency
 * injection) process regardless of whether the {@link Configurable} interface
 * is implemented.
 * </p>
 * <p>
 * At this point in the process any bean properties are set. A bean property may
 * be of any type that can also be a constructor arg (although if it is a
 * reference to another bean, a different construct is used -- see below).
 * Additionally, properties can be maps or lists.
 * </p>
 * <p>
 * The simple form for setting a property [note that the property's type for a
 * simple property -- not a list or map -- is entirely determined by the
 * signature of the bean's <code>setX()</code> method] is:
 * </p>
 * 
 * <pre>
 * &lt;property name=&quot;property1&quot;&gt;&lt;value&gt;1&lt;/value&gt;&lt;/property&gt;
 * </pre>
 * <p>
 * The construct used to set a property which is a reference to another bean is:
 * </p>
 * 
 * <pre>
 * &lt;property name=&quot;property1&quot;&gt;&lt;value&gt;!My Bean&lt;/value&gt;&lt;/property&gt;
 * </pre>
 * <p>
 * In this case it doesn't matter where in the XML file the referenced bean is
 * defined. If the string you want really begins with an exclamation mark, you
 * must give two exclamation marks -- which will be converted to one, literal,
 * exclamation mark.
 * </p>
 * <p>
 * For more details on configuring bean properties, please refer to the javadoc
 * for {@link BeanMap}.
 * </p>
 * <p>
 * Finally, beans which implement the {@link Runnable} interface will be run
 * when the {@link #run()} method is called on the container.
 * 
 * For exemplars, see the {@link #main(String[])} program herein, or look at the
 * unit tests.
 * 
 * In addition to the beans explicitly defined in the list of beans, the BeanPot
 * manages two "agents": the "writer", which is always available, and the
 * "MBeanPot" which is optional. The writer is referenced when getting a bean as
 * "_Writer_".
 * </p>
 * <p>
 * Three types of beans are supported:
 * </p>
 * <dl>
 * <dt>instance</dt>
 * <dd>the default type of bean which is instantiated by the bean pot using a
 * public constructor (according to the constructor arguments provided);</dd>
 * <dt>singleton</dt>
 * <dd>a bean which is the sole instance of a singleton class, referenced by the
 * bean pot by invoking the class method <i>getInstance()</i> on the class;</dd>
 * <dt>external</dt>
 * <dd>a bean which is externally imposed on the container before configuration
 * using the method {@link BeanPot#imposeBean(String, Object)} [the bean
 * container is not responsible for creating the reference to the bean as it is
 * for the instance and singleton types].</dd>
 * </dl>
 * <p>
 * Note that we use a simple logger for internal logging (it saves having an
 * extra dependency on a logging technology). The BeanMap logger can be set to
 * whatever is required. It extends {@link java.util.logging.Logger}.
 * 
 * The beans are stored in an instance of {@link BeanMap}. For more details on
 * configuring constructor arguments and properties of beans, please refer to
 * the javadoc for {@link BeanMap}.
 * </p>
 * <p>
 * A useful feature allows you to configure a bean as a member of another bean.
 * Thus, once the new bean is added to the bean map successfully, it is also
 * added as a member of the referenced bean. There are two forms of membership:
 * keyed membership (corresponding to a map) and unkeyed membership
 * (corresponding to a list). The syntax is as in the following examples.
 * </p>
 * 
 * <pre>
 * &lt;bean id=&quot;bean1&quot; class=&quot;com....&quot;&gt;
 *     &lt;member membership=&quot;child&quot;&gt;!bean2&lt;/member&gt;
 * &lt;/bean&gt;
 * </pre>
 * <p>
 * which adds bean1 as a child of bean2 by invoking the method addChild() on
 * bean2 with bean1 as the parameter.<br/>
 * Alternatively,
 * </p>
 * 
 * <pre>
 * &lt;bean id=&quot;bean1&quot; class=&quot;com....&quot;&gt;
 *     &lt;member key="x" membership=&quot;child&quot;&gt;!bean2&lt;/member&gt;
 * &lt;/bean&gt;
 * </pre>
 * <p>
 * which adds bean1 as a child of bean2 by invoking the method putChild() on
 * bean2 with parameters "x" and bean1.
 * </p>
 * 
 * @author Robin Hillyard
 * 
 *         TODO consider using the features provided by the {@link BeanFactory}
 *         and its related types.
 */
@SuppressWarnings("unchecked")
public final class BeanPot implements ConfigurationBuilder, Runnable {

	/**
	 * Public class method to add a bean to the bean pot.
	 * 
	 * @param key
	 *            the identifier for the new bean
	 * @param bean
	 *            the bean to be added
	 * @param registerAsMBean
	 *            if true or maybe, we try to register the bean as an MBean
	 * @return the previous bean, if we're replacing (else null).
	 * @throws BeanPotException
	 * @see BeanMap#addBean(String, Object, Kleenean)
	 */
	public static final Object addBean(final String key, final Object bean, final Kleenean registerAsMBean)
			throws BeanPotException {
		return getInstance().getBeanMap().addBean(key, bean, registerAsMBean);
	}

	/**
	 * Remove the beans from the bean map. Additionally, any MBeans will be
	 * unregistered. Furthermore, the beanpot's configuration will be voided.
	 * 
	 * @return
	 */
	public final static boolean cleanup() {
		final BeanPot beanpot = BeanPot.getInstance();
		final boolean ok = beanpot.getBeanMap().cleanup();
		beanpot.doSetConfigDetails(NULL_CONFIGURATION);
		setValidate(false);
		return ok;
	}

	/**
	 * Public class method to configure the bean pot with nulls passed to the
	 * preConfig and postConfig steps.
	 * 
	 * @throws BeanPotException
	 * 
	 * @see #doConfigure()
	 */
	public static final void configure() throws BeanPotException {
		getInstance().doConfigure();
	}

	/**
	 * Public class method to configure the bean pot. It is optional for the
	 * bean pot application to invoke this method explicitly, assuming that the
	 * application invokes the {@link #run()} method. If the {@link #run()}
	 * method is invoked before this method, then this method will be invoked
	 * automatically. Additionally, if the bean pot has been configured already,
	 * we don't do it again.
	 * 
	 * The steps are as follows:
	 * <ol>
	 * <li>read the configuration and for each bean defined, construct the bean
	 * with the no-arg constructor;</li>
	 * <li>call {@link #preConfigureBeans(Object)} passing
	 * <code>preConfigData</code> as parameter</li>
	 * <li>for each bean, configure it according to its entry in the
	 * configuration, doing dependency injection as we go;</li>
	 * <li>call {@link #postConfigureBeans(Object)} passing
	 * <code>postConfigData</code> as parameter</li>
	 * </ol>
	 * 
	 * @param preData
	 *            data to be passed to the
	 *            {@link Configurable#preConfigure(Object)} method of each
	 *            {@link Configurable} bean
	 * @param postData
	 *            data to be passed to the
	 *            {@link Configurable#postConfigure(Object)} method of each
	 *            {@link Configurable} bean
	 * @throws BeanPotException
	 */
	public static final void configure(final Object preData, final Object postData) throws BeanPotException {
		getInstance().doConfigure(preData, postData);
	}

	/**
	 * Method to find the number of beans currently under management.
	 * 
	 * @return the current number of beans in this container
	 */
	public static final int count() {
		return getInstance().getBeanMap().size();
	}

	/**
	 * Get the bean given the key (its name). If the bean is not found, a
	 * warning will be sent to the writer.
	 * 
	 * @param key
	 * @return the bean else null (if no such bean).
	 * @see Map#get(Object)
	 */
	public static final Object getBean(final String key) {
		if (key.equals(AGENT_WRITER_))
			return getWriter();
		if (key.equals(AGENT_MBEAN_POT_))
			return getMBeanPot();
		return getInstance().getBeanMap().getBean(key);
	}

	/**
	 * Public class method to get the set of keys which index the beans.
	 * 
	 * @return the set of Strings which identify the various beans
	 */
	public static final Set<String> getBeanKeys() {
		return getInstance().getBeanMap().getBeanKeys();
	}

	/**
	 * Method to get the logger.
	 * 
	 * @return the logger for this Class.
	 */
	public static final Log getBeanPotLogger() {
		if (beanPotLogger == null)
			beanPotLogger = LogFactory.getLog(BeanPot.class);

		return beanPotLogger;
	}

	/**
	 * @return the dTD
	 */
	public static final URL getDTD() {
		return DTD;
	}

	/**
	 * @return the {@link #identifier}
	 */
	public static final String getIdentifier() {
		return getInstance().identifier;
	}

	/**
	 * Class method to get the (singleton) instance of BeanPot.
	 * 
	 * @return the (singleton) BeanPot.
	 */
	public static final BeanPot getInstance() {
		if (Instance == null)
			Instance = new BeanPot();
		return Instance;
	}

	/**
	 * @return the current value of the class field MBeanServer.
	 */
	public final static Object getMBeanServer() {
		return MBeanServer;
	}

	/**
	 * method to get the current pattern to match in the configuration file for
	 * finding the beans.
	 * 
	 * @return the currently defined (static) pattern for extracting beans from
	 *         the configuration file.
	 * @see #Pattern
	 */
	public static final String getPattern() {
		return Pattern;
	}

	/**
	 * Get the writer for the bean pot.
	 * 
	 * TEST
	 * 
	 * @return the instance of PrintWriter which has been defined for this
	 *         container.
	 * 
	 *         XXX note that we are actually setting a class value here, which
	 *         is OK since this is a singleton class, but not really great.
	 */
	public static final PrintWriter getWriter() {
		if (getBeanPotLogger() instanceof BeanPotLogger) {
			return ((BeanPotLogger) getBeanPotLogger()).getWriter();
		}
		return printWriter;
	}

	/**
	 * The key should correspond to the name of a bean of type
	 * {@link Type#EXTERNAL} which will be defined in the configuration file.
	 * 
	 * @param key
	 *            the key by which the bean will be reference.
	 * @param bean
	 *            the bean being externally imposed.
	 * @throws BeanPotException
	 */
	public static final void imposeBean(final String key, final Object bean) throws BeanPotException {
		getInstance().getBeanMap().doImposeBean(key, bean);
	}

	/**
	 * Public class method to initialize the bean pot. It is optional for the
	 * container application to invoke this method explicitly, assuming that the
	 * application invokes the {@link #doConfigure()} method. If the
	 * {@link #doConfigure()} method is invoked before this method, then this
	 * method will be invoked automatically. Additionally, if the bean pot has
	 * been initialized already, we don't do it again.
	 * 
	 * The steps are as follows:
	 * <ol>
	 * <li>configure the writer (optional);</li>
	 * <li>configure MBean server;</li>
	 * <li>configure the imposed bean (optional);</li>
	 * <li>set initialized as true.</li>
	 * </ol>
	 * 
	 * @throws BeanPotException
	 */
	public static final void init() throws BeanPotException {
		getInstance().doInit();
	}

	/**
	 * @return the validate
	 */
	public static boolean isValidate() {
		return BeanPot.validate;
	}

	/**
	 * This is here essentially only for testing (in addition to the unit
	 * tests). You can also use this code fragment as an exemplar. By default,
	 * this main program looks for a file called beans.xml in the resource
	 * package corresponding to this class BeanPot. In my set up, this means
	 * it's in my workarea at:
	 * <code>beanpot/src/main/resource/com/rubecula/beanpot/beans.xml</code> You
	 * can override the location by providing a parameter to the argument list.
	 * 
	 * @param args
	 *            only one argument is consulted -- see above
	 */
	public final static void main(final String[] args) {
		String beansFilename = "beans.xml"; //$NON-NLS-1$
		if (args.length > 0)
			beansFilename = args[0];
		try {
			BeanPot.setValidate();
			BeanPot.setConfiguration(BeanPot.class, beansFilename).run();
			// count should be 1
			BeanPot.getWriter().println("BeanPot running with " + BeanPot.count() + " beans."); //$NON-NLS-1$//$NON-NLS-2$
			final String integerBeanKey = "integer"; //$NON-NLS-1$
			final Integer bean = (Integer) BeanPot.getBean(integerBeanKey);
			if (bean != null) {
				// integer value should be 99
				BeanPot.getWriter().println(integerBeanKey + " value: " + bean.intValue()); //$NON-NLS-1$
			} else
				System.err.println("cannot find bean " + integerBeanKey); //$NON-NLS-1$
		} catch (final Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * @param file
	 *            a {@link File} reference, which should exist beforehand,
	 *            having been filled with serialized objects.
	 * @param keys
	 * @param deleteAfter
	 *            XXX
	 * @throws BeanPotException
	 * @throws IOException
	 * 
	 * 
	 */
	public static void restoreBeans(final File file, final String[] keys, final boolean deleteAfter) throws IOException,
			BeanPotException {
		getInstance().restoreBeans(SerializerFactory.createDeserializer(file, deleteAfter), keys);
	}

	/**
	 * @param filename
	 *            a {@link File} reference, which should exist beforehand,
	 *            having been filled with serialized objects.
	 * @param keys
	 * @param deleteAfter
	 *            XXX
	 * @throws BeanPotException
	 * @throws IOException
	 * 
	 * 
	 */
	public static void restoreBeans(final String filename, final String[] keys, final boolean deleteAfter) throws IOException,
			BeanPotException {
		getInstance().restoreBeans(SerializerFactory.createDeserializer(filename, deleteAfter), keys);
	}

	/**
	 * @param file
	 *            a {@link File} reference, which should not exist beforehand.
	 * @return an array of bean keys
	 * @throws IOException
	 * 
	 * 
	 */
	public static String[] serializeBeans(final File file) throws IOException {
		return getInstance().serializeBeans(SerializerFactory.createSerializer(file));
	}

	/**
	 * @param filename
	 *            a {@link File} reference, which should not exist beforehand.
	 * @return an array of bean keys
	 * @throws IOException
	 * 
	 * 
	 */
	public static String[] serializeBeans(final String filename) throws IOException {
		return getInstance().serializeBeans(SerializerFactory.createSerializer(filename));
	}

	/**
	 * @param beanPropertyName
	 * @param value
	 * @return the bean which has been updated.
	 * @throws BeanPotException
	 * 
	 *             XXX should this be final?
	 */
	public static Object setBeanProperty(final String beanPropertyName, final Object value) throws BeanPotException {
		final Matcher matcher = BeanMap.getBeanPropertyPattern().matcher(beanPropertyName);
		if (matcher.matches()) {
			if (matcher.groupCount() == 2)
				return getInstance().getBeanMap().setBeanProperty(matcher.group(1), matcher.group(2), value);
			throw new BeanPotException("invalid bean property name: " + beanPropertyName); //$NON-NLS-1$
		}
		throw new BeanPotException("invalid bean property name: " + beanPropertyName); //$NON-NLS-1$
	}

	/**
	 * Set the class loader which will be used to satisfy the bean definitions.
	 * If this method is not called, the default class loader will be used
	 * (that's to say the class loader which loads the {@link BeanMap} class).
	 * 
	 * @param classLoader
	 */
	public static final void setClassLoader(final ClassLoader classLoader) {
		getInstance().getBeanMap().setClassLoader(classLoader);
	}

	/**
	 * Set configuration by resource.
	 * 
	 * XXX inconsistent use of return value. Other similar methods don't follow
	 * the same convention.
	 * 
	 * @param clazz
	 * @param filename
	 * @return the BeanPot with configuration set.
	 * @throws BeanPotException
	 */
	public final static BeanPot setConfiguration(final Class<? extends Object> clazz, final String filename)
			throws BeanPotException {
		final URL url = clazz.getResource(filename);
		if (url != null)
			return setConfiguration(url);
		throw new BeanPotException("cannot find resource for filename: " + filename + " relative to " + clazz); //$NON-NLS-1$
	}

	/**
	 * Public class method to set the configuration as given. The configuration
	 * must be set before the BeanPot can be initialized.
	 * 
	 * @param configFile
	 *            a (XML) File representing the configuration.
	 * @return this (as a convenience).
	 * @throws BeanPotException
	 */
	public static final BeanPot setConfiguration(final File configFile) throws BeanPotException {
		return getInstance().doSetConfiguration(configFile);
	}

	/**
	 * Public class method to set the configuration as given. The configuration
	 * must be set before the BeanPot can be initialized.
	 * 
	 * If the configuration has been previously set, a warning is logged.
	 * 
	 * In any event, the bean map is cleared and the initialized flag is set to
	 * false.
	 * 
	 * @param config
	 *            the configuration, normally derived from an XML file.
	 * @return this (as a convenience).
	 */
	public static final BeanPot setConfiguration(final HierarchicalConfiguration config) {
		return getInstance().doSetConfiguration(config);
	}

	/**
	 * Public class method to set the configuration as given. The configuration
	 * must be set before the BeanPot can be initialized.
	 * 
	 * XXX this is never called.
	 * 
	 * @param configFile
	 *            the pathname of the configuration file (in XML).
	 * @return this (as a convenience).
	 * @throws BeanPotException
	 * @see #setConfiguration(File)
	 */
	public static final BeanPot setConfiguration(final String configFile) throws BeanPotException {
		return getInstance().doSetConfiguration(configFile);
	}

	/**
	 * Public class method to set the configuration as given. The configuration
	 * must be set before the BeanPot can be initialized.
	 * 
	 * @param url
	 *            URL which points to an XML file representing the
	 *            configuration.
	 * @return this (as a convenience).
	 * @throws BeanPotException
	 */
	public static final BeanPot setConfiguration(final URL url) throws BeanPotException {
		return getInstance().doSetConfiguration(url);
	}

	/**
	 * Set the debug flag.
	 * 
	 * @param debug
	 *            whether we want debug output or not.
	 */
	public static final void setDebug(final boolean debug) {
		final Log log = getBeanPotLogger();
		if (log instanceof BeanPotLogger) {
			((BeanPotLogger) log).setDebug(debug);
		} else if (log instanceof java.util.logging.Logger) {
			final java.util.logging.Logger logger = (java.util.logging.Logger) log;
			logger.setLevel(Level.FINE);
		} else
			log.info("BeanPot.setDebug(): can't set debug on this logger of " + log.getClass() + " -- consider setting the commons log factory to " + BeanPotLogFactory.class.getName()); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * This method allows an application to set the MBeanServer to use
	 * explicitly. In order to be effective, it must be called before the
	 * {@link #init()} method is called.
	 * 
	 * @param beanServer
	 *            an object whose type is javax.management.MBeanServer
	 */
	public final static void setMBeanServer(final Object beanServer) {
		MBeanServer = beanServer;
	}

	/**
	 * Set a different pattern for extracting beans from the configuration file.
	 * 
	 * @param pattern
	 *            the new pattern.
	 * 
	 * @see #Pattern default pattern is the value of {@link #S_BEANS_BEAN}.
	 */
	public static final void setPattern(final String pattern) {
		BeanPot.Pattern = pattern;
	}

	/**
	 * @param postConfigData
	 * 
	 *            XXX should this be final?
	 */
	public static void setPostConfigData(final Object postConfigData) {
		BeanPot.postConfigData = postConfigData;
	}

	/**
	 * @param preConfigData
	 * 
	 *            XXX should this be final?
	 */
	public static void setPreConfigData(final Object preConfigData) {
		BeanPot.preConfigData = preConfigData;
	}

	/**
	 * Set the validating flag to true. Invokes {@link #setValidate()} with
	 * parameter <code>true</code>.
	 */
	public static void setValidate() {
		setValidate(true);
	}

	/**
	 * Set the validating flag. Note that this is a class method as opposed to
	 * {@link BeanConfiguration#setValidating(boolean)} which is an instance
	 * method.
	 * 
	 * @param validate
	 */
	public static void setValidate(final boolean validate) {
		BeanPot.validate = validate;

	}

	/**
	 * Set the validating flag and set the URL for the DTD. Invoke
	 * {@link #setValidate(boolean)} and {@link #setDTD(URL)}.
	 * 
	 * @param validate
	 * @param dtd
	 *            the URL for the DTD (if null then we stick with the current
	 *            value).
	 */
	public static void setValidate(final boolean validate, final URL dtd) {
		setValidate(validate);
		if (dtd != null)
			setDTD(dtd);
	}

	/**
	 * Set the validating flag and set the URL for the DTD. Invoke
	 * {@link #setValidate(boolean)} with value <code>true</code> and
	 * {@link #setDTD(URL)}.
	 * 
	 * @param dtd
	 *            the URL for the DTD
	 */
	public static void setValidate(final URL dtd) {
		setValidate(true, dtd);
	}

	/**
	 * 
	 * 
	 * XXX should this be final?
	 */
	public static void showBeans() {
		getInstance().showBeans(getWriter());
	}

	/**
	 * @return the current value of {@link #MBeanPot}.
	 */
	final static MBeanPot getMBeanPot() {
		return MBeanPot;
	}

	/**
	 * Remove the bean identified by key.
	 * 
	 * @param key
	 *            the key to identify the bean (usually its name).
	 * @return the former bean corresponding to key, else null
	 * 
	 * @see Map#remove(Object)
	 */
	static final Object removeBean(final String key) {
		return getInstance().getBeanMap().removeBean(key);
	}

	/**
	 * @param url
	 *            the URL of the DTD to use for validation
	 */
	static final void setDTD(final URL url) {
		DTD = url;
	}

	/**
	 * @param mBeanPot
	 */
	static final void setMBeanPot(final MBeanPot mBeanPot) {
		MBeanPot = mBeanPot;
	}

	/**
	 * Set the {@link PrintWriter} for this container (used for warnings and
	 * debugging).
	 * 
	 * @param writer
	 *            a {@link PrintWriter} instance.
	 * 
	 *            XXX note that we are actually setting a class value here,
	 *            which is OK since this is a singleton class, but not really
	 *            great.
	 */
	static final void setWriter(final PrintWriter writer) {
		printWriter = writer;
		// The bean pot logger shares the same writer (assuming it's a
		// BeanPotLogger).
		if (getBeanPotLogger() instanceof BeanPotLogger) {
			((BeanPotLogger) getBeanPotLogger()).setWriter(printWriter);
		}
	}

	/**
	 * Method to construct an agent for the BeanPot. An agent is a bean in the
	 * sense that it is constructed like a bean, but it does not go into the
	 * bean map.
	 * 
	 * @param id
	 * @param agentConfig
	 * @param constructorArgs
	 *            constructor args
	 * @return the bean which will be used for the writer.
	 * 
	 * @throws BeanPotException
	 */
	private static Object constructAgent(final String id, final Configuration agentConfig, final Object[] constructorArgs)
			throws BeanPotException {
		String agentClass = agentConfig.getString(BeanUtilities.S_ATTR_CLASS);
		if (agentClass == null)
			agentClass = lookupAgentClassDefault(id);
		getBeanPotLogger().debug("Constructing agent: " + id + " with class: " + agentClass); //$NON-NLS-1$ //$NON-NLS-2$
		if (agentClass != null)
			return BeanUtilities.constructBean(agentClass, BeanPot.class.getClassLoader(), id, constructorArgs, null);
		throw new BeanPotException("class attribute is missing for " + id); //$NON-NLS-1$
	}

	/**
	 * @param id
	 * @return
	 */
	private static String lookupAgentClassDefault(final String id) {
		return _agentClassDefaults.get(id);
	}

	/**
	 * Private Primary Constructor.
	 */
	private BeanPot() {
		super();
		beanPotLogger = LogFactory.getLog(BeanPot.class);
		_agentClassDefaults.put(S_WRITER, "com.rubecula.beanpot.ConsoleWriter"); //$NON-NLS-1$
		_agentClassDefaults.put(S_MBEAN_POT, "com.rubecula.beanpot.jmx.MBeanPotImpl"); //$NON-NLS-1$
	}

	/**
	 * Get the configuration as a {@link Configuration} object.
	 * 
	 * This is an instance method because BeanPot implements the
	 * {@link ConfigurationBuilder} interface.
	 * 
	 * @see org.apache.commons.configuration.ConfigurationBuilder#getConfiguration()
	 */
	public final Configuration getConfiguration() throws ConfigurationException {
		if (getConfig() != null)
			return getConfig();
		throw new ConfigurationException("configuration is null"); //$NON-NLS-1$
	}

	/**
	 * Serialize all of the {@link Serializable} beans into the given
	 * <code>serializer</code>.
	 * 
	 * TODO need to restore MBeans also.
	 * 
	 * @param serializer
	 *            a serializer (typically a File) which will hold the serialized
	 *            beans.
	 * @param keys
	 *            XXX
	 * @throws IOException
	 * @throws BeanPotException
	 */
	public final void restoreBeans(final Serializer serializer, final String[] keys) throws IOException, BeanPotException {
		for (int i = 0; i < keys.length; i++)
			restoreBean(serializer, getBeanMap(), keys[i]);
		if (serializer.available() > 0)
			beanPotLogger.warn("restoreBeans(): more data available"); //$NON-NLS-1$
		serializer.close();
	}

	/**
	 * This instance method "runs" the BeanPot (container). The steps are as
	 * follows:
	 * <ol>
	 * <li>if the bean pot is not configured, the we call {@link #doConfigure()}
	 * ;</li>
	 * <li>create a new thread on the for each Runnable bean and call its
	 * {@link Thread#start()} method.</li>
	 * </ol>
	 * 
	 * Note that this is an instance method because BeanPot implements the
	 * Runnable interface.
	 * 
	 * @see java.lang.Runnable#run()
	 * @throws RuntimeException
	 */
	public final void run() {

		if (!isConfigured()) {
			try {
				doConfigure();
			} catch (final BeanPotException e) {
				throw new RuntimeException("BeanPot.run(): exception: " + e.getLocalizedMessage(), e.getCause()); //$NON-NLS-1$
			}
		}

		getBeanPotLogger().debug("Running runnable beans in container"); //$NON-NLS-1$

		getBeanMap().runBeans();
	}

	/**
	 * Serialize all of the {@link Serializable} beans into the given
	 * <code>serializer</code>.
	 * 
	 * @param serializer
	 *            a {@link Serializer} which will accept the serialized beans.
	 * @return an array of bean keys which have been serialized (in order).
	 * @throws IOException
	 */
	public final String[] serializeBeans(final Serializer serializer) throws IOException {
		final Collection<String> result = new ArrayList<String>();
		for (final String key : new TreeSet<String>(getBeanMap().getBeanKeys())) {
			final Object bean = getBean(key);
			if (bean instanceof Serializable) {
				serializer.writeObject(bean);
				result.add(key);
			}
		}
		serializer.close();
		return result.toArray(new String[] {});
	}

	/**
	 * @param writer
	 */
	public final void showBeans(final PrintWriter writer) {
		writer.println("Bean pot identifier: " + getIdentifier()); //$NON-NLS-1$
		if (getMBeanPot() != null)
			writer.println("  with MBeanPot"); //$NON-NLS-1$
		for (final String key : new TreeSet<String>(getBeanMap().getBeanKeys())) {
			showBean(key, writer);
		}
	}

	/**
	 * Optionally set the MBeanPot agent according to the bean specified in the
	 * configuration
	 * 
	 * @throws BeanPotException
	 */
	final void configureMBeanServer() throws BeanPotException {
		final Object agent = configureAgent(getConfig(), S_MBEAN_POT);
		if (agent instanceof MBeanPot) {
			setMBeanPot((MBeanPot) agent);
		}
	}

	/**
	 * Add the beans to the bean map, according to the configurations.
	 * 
	 * @param configurations
	 * @throws BeanPotException
	 */
	private void addBeans(final Collection<HierarchicalConfiguration> configurations) throws BeanPotException {
		for (final HierarchicalConfiguration configuration : configurations) {
			showConfig(configuration);
			getBeanMap().addBean(configuration);
		}
	}

	/**
	 * This method adds any beans that are defined in <code>configuration</code>
	 * . Note that, because the {@link XMLConfiguration} class does not support
	 * includes, we have to do it ourselves. Anything of the form
	 * &lt;include&gt;value&lt;/include&gt; will be processed as an include. The
	 * value will be considered relative to the working directory first, and if
	 * that fails then relative to the class in which the configuration resource
	 * was found.
	 * 
	 * This method is expected to be called recursively.
	 * 
	 * @param parentConfig
	 * @return a collection of hierarchical configurations corresponding to
	 *         beans.
	 * @throws BeanPotException
	 */
	private Collection<HierarchicalConfiguration> addBeans(final HierarchicalConfiguration parentConfig) throws BeanPotException {
		final Collection<HierarchicalConfiguration> beanConfigurations = parentConfig.configurationsAt(getPattern());
		addBeans(beanConfigurations);
		for (final HierarchicalConfiguration include : (Collection<HierarchicalConfiguration>) parentConfig
				.configurationsAt(PATTERN_INCLUDE))
			beanConfigurations.addAll(processIncludeElement(parentConfig, include));
		return beanConfigurations;
	}

	/**
	 * Configure an agent for this BeanPot.
	 * 
	 * @param configuration
	 *            the configuration typically from {@link XMLConfiguration}.
	 * @param tag
	 *            the string used for the element in the XML file to define the
	 *            agent.
	 * @return the agent (or null)
	 * @throws BeanPotException
	 */
	private Object configureAgent(final HierarchicalConfiguration configuration, final String tag) throws BeanPotException {
		if (configuration.configurationsAt(tag).size() == 1) {
			final SubnodeConfiguration agentConfiguration = configuration.configurationAt(tag);
			if (agentConfiguration != null) {
				if (tag.equalsIgnoreCase(S_WRITER))
					return BeanPot.constructAgent(tag, agentConfiguration, new Object[0]);
				else if (tag.equalsIgnoreCase(S_MBEAN_POT))
					return constructMBeanPot(tag, agentConfiguration);
				// else if (tag.equalsIgnoreCase(S_IMPOSED_BEAN))
				// configureImposedBean(agentConfiguration);
				else if (tag.equalsIgnoreCase(S_IDENTIFIER))
					configureIdentifier(agentConfiguration);
				else
					throw new BeanPotException("configureAgent(): unknown tag: " + tag); //$NON-NLS-1$
			}
			return null;
		}
		return null;
	}

	/**
	 * Configure the beans in the bean map according to the configurations.
	 * 
	 * @param configurations
	 * @throws BeanPotException
	 */
	private void configureBeans(final Collection<HierarchicalConfiguration> configurations) throws BeanPotException {
		boolean ok = true;
		for (final HierarchicalConfiguration configuration : configurations) {
			try {
				getBeanMap().configureBean(configuration);
			} catch (final Exception e) {
				ok = false;
				getBeanPotLogger().error("configuration error", e); //$NON-NLS-1$
			}
		}
		if (!ok)
			throw new BeanPotException("configuration failed (see previous messages)"); //$NON-NLS-1$
	}

	/**
	 * @throws BeanPotException
	 * 
	 */
	private void configureIdentifier() throws BeanPotException {
		configureAgent(getConfig(), S_IDENTIFIER);
	}

	/**
	 * @param agentConfiguration
	 */
	private void configureIdentifier(final SubnodeConfiguration agentConfiguration) {
		final Iterator<String> keys = agentConfiguration.getKeys();
		while (keys.hasNext()) {
			final String key = keys.next();
			final String value = agentConfiguration.getString(key);
			if (key.equalsIgnoreCase("")) //$NON-NLS-1$
				setIdentifier(value);
		}

	}

	/**
	 * Optionally set the writer according to the class specified in the
	 * configuration
	 * 
	 * @throws BeanPotException
	 */
	private void configureWriter() throws BeanPotException {
		final Object writer = configureAgent(getConfig(), S_WRITER);
		if (writer != null) {
			if (writer instanceof PrintWriter)
				setWriter((PrintWriter) writer);
			else
				throw new BeanPotException("class for " + S_WRITER + " does not extend PrintWriter"); //$NON-NLS-1$ //$NON-NLS-2$
		}
	}

	/**
	 * Method which first checks to see if the MBeanServer field has been set
	 * previously by {@link #setMBeanServer(Object)}, in which case it must be
	 * of type javax.management.MBeanServer. If not, it will be set to the
	 * platform mbean server: {@link ManagementFactory#getPlatformMBeanServer()}
	 * . Finally, the mbean server is passed as the constructor of the MBeanPot
	 * class.
	 * 
	 * @param tag
	 * @param agentConfiguration
	 * @return the newly constructed MBeanPot
	 * @throws BeanPotException
	 */
	private Object constructMBeanPot(final String tag, final SubnodeConfiguration agentConfiguration) throws BeanPotException {
		if (getMBeanServer() == null) {
			setMBeanServer(ManagementFactory.getPlatformMBeanServer());
		}
		final String prefix = agentConfiguration.getString(S_ATTR_PREFIX);
		if (prefix != null)
			MBeanPrefix = prefix;

		// Note that the getMBeanServer() result must be of type
		// javax.management.MBeanServer
		return BeanPot.constructAgent(tag, agentConfiguration, new Object[] { getMBeanServer(), MBeanPrefix });
	}

	/**
	 * instance method to configure with nulls passed to the preConfig and
	 * postConfig steps.
	 * 
	 * @throws BeanPotException
	 * 
	 * @see #configure(Object, Object)
	 */
	private void doConfigure() throws BeanPotException {
		configure(preConfigData, postConfigData);
	}

	/**
	 * Private instance method to configure this bean pot. For description, see
	 * {@link #configure(Object, Object)}.
	 * 
	 * @param preData
	 *            data to be passed to the
	 *            {@link Configurable#preConfigure(Object)} method of each
	 *            {@link Configurable} bean
	 * @param postData
	 *            data to be passed to the
	 *            {@link Configurable#postConfigure(Object)} method of each
	 *            {@link Configurable} bean
	 * @throws BeanPotException
	 */
	private void doConfigure(final Object preData, final Object postData) throws BeanPotException {
		if (isConfigured())
			return;

		if (!isInitialized()) {
			init();
		}

		if (getConfig() != null) {
			final Collection<HierarchicalConfiguration> beanConfigurations = addBeans(getConfig());
			getBeanMap().configureBeanMapProperties(getConfig());
			preConfigureBeans(preData);
			configureBeans(beanConfigurations);
			postConfigureBeans(postData);

			setConfigured(true);

			getBeanPotLogger().debug("Container is configured with " + count() + " beans"); //$NON-NLS-1$ //$NON-NLS-2$
		} else
			throw new BeanPotException("BeanPot.configure(Object,Object): configuration has not been set"); //$NON-NLS-1$
	}

	/**
	 * Private instance method to do the initialization of this bean pot. For
	 * description, see {@link #init()}.
	 * 
	 * @throws BeanPotException
	 */
	private void doInit() throws BeanPotException {
		if (isInitialized())
			return;

		if (getConfig() != null) {
			configureWriter();

			configureIdentifier();

			// XXX consider putting this before configureWriter()
			getBeanPotLogger().debug("BeanPot: initializing container"); //$NON-NLS-1$

			if (getConfigurationSource() != null)
				getBeanPotLogger().debug("BeanPot: configuration source: " + getConfigurationSource()); //$NON-NLS-1$

			configureMBeanServer();

			// configureImposedBean();

			setInitialized(true);

		} else
			throw new BeanPotException("BeanPot.init(Object,Object): configuration has not been set"); //$NON-NLS-1$
	}

	/**
	 * @param configuration
	 *            may be null
	 */
	private void doSetConfigDetails(final HierarchicalConfiguration configuration) {
		setConfig(configuration);
		getBeanMap().clear();
		setInitialized(false);
		setConfigured(false);
	}

	/**
	 * Private instance method to support {@link #setConfiguration(File)}.
	 * 
	 * @param configFile
	 * @return this
	 * @throws BeanPotException
	 */
	private BeanPot doSetConfiguration(final File configFile) throws BeanPotException {
		if (configFile != null) {
			final boolean exists = configFile.exists();
			if (exists) {
				try {
					return setConfiguration(BeanConfiguration.create(configFile));
				} catch (final ConfigurationException e) {
					throw new BeanPotException(S_M_SET_CONFIGURATION_FILE + S_COLON + configFile.getAbsolutePath(), e);
				} finally {
					// do nothing
				}
			}
			throw new BeanPotException(S_M_SET_CONFIGURATION_FILE + S_COLON + configFile.getAbsolutePath() + " does not exist"); //$NON-NLS-1$
		}
		throw new BeanPotException(S_M_SET_CONFIGURATION_FILE + S_COLON + "configFile is null"); //$NON-NLS-1$
	}

	/**
	 * Private instance method to set the configuration of this bean pot as
	 * given. The configuration must be set before the BeanPot can be
	 * initialized.
	 * 
	 * Note that here we simply <i>set</i> the configuration... we do not
	 * <i>do</i> the configuration.
	 * 
	 * For details, see {@link #setConfiguration(HierarchicalConfiguration)}.
	 * 
	 * @param configuration
	 *            the configuration, normally derived from an XML file.
	 * @return this (as a convenience).
	 */
	private BeanPot doSetConfiguration(final HierarchicalConfiguration configuration) {
		if (getConfig() != null && configuration != getConfig())
			getBeanPotLogger().warn("BeanPot.setConfiguration(): bean pot configuration has already been set"); //$NON-NLS-1$
		doSetConfigDetails(configuration);
		return this;
	}

	/**
	 * Private instance method to set the configuration of this bean pot as
	 * given.
	 * 
	 * For details, see {@link #setConfiguration(String)}
	 * 
	 * @param configFile
	 *            the pathname of the configuration file (in XML).
	 * @return this (as a convenience).
	 * @throws BeanPotException
	 * @see #setConfiguration(File)
	 */
	private BeanPot doSetConfiguration(final String configFile) throws BeanPotException {
		setConfigurationSource("file: " + configFile); //$NON-NLS-1$
		setConfiguration(new File(configFile));
		return this;
	}

	/**
	 * Private instance method to set the configuration of this bean pot as
	 * given.
	 * 
	 * For details, see {@link #setConfiguration(URL)}.
	 * 
	 * @param url
	 *            URL which points to an XML file representing the
	 *            configuration.
	 * @return this (as a convenience).
	 * @throws BeanPotException
	 */
	private BeanPot doSetConfiguration(final URL url) throws BeanPotException {
		if (url != null) {
			try {
				setConfigurationSource("url: " + url); //$NON-NLS-1$
				return doSetConfiguration(BeanConfiguration.create(url));
			} catch (final ConfigurationException e) {
				throw new BeanPotException("setConfiguration(URL) configuration exception: " + url, e); //$NON-NLS-1$
			}
		}
		throw new BeanPotException("setConfiguration(URL): url does not exist"); //$NON-NLS-1$
	}

	/**
	 * @return the bean map for this bean pot
	 */
	private BeanMap getBeanMap() {
		return this._beanMap;
	}

	private HierarchicalConfiguration getConfig() {
		return this.config;
	}

	/**
	 * @return the configurationSource
	 */
	private String getConfigurationSource() {
		return this.configurationSource;
	}

	/**
	 * @return true if this bean pot has already been configured.
	 */
	private boolean isConfigured() {
		return this.configured;
	}

	/**
	 * method to determine if this bean pot has been initialized.
	 * 
	 * @return true if the beans have been initialized (configured).
	 */
	private boolean isInitialized() {
		return this.initialized;
	}

	/**
	 * Method to invoke {@link Configurable#postConfigure(Object)} on each bean
	 * in the container, with data passed in as the parameter.
	 * 
	 * @param data
	 *            arbitrary data for post-configuration.
	 * @throws BeanPotException
	 * @see Configurable#postConfigure(Object)
	 */
	private void postConfigureBeans(final Object data) throws BeanPotException {
		getBeanPotLogger().debug("Post-configure beans in container"); //$NON-NLS-1$
		boolean ok = true;
		for (final String key : getBeanKeys()) {
			final Object bean = getBean(key);
			// TODO consider doing this via polymorphism
			if (bean instanceof Configurable)
				try {
					((Configurable) bean).postConfigure(data);
				} catch (final BeanPotException e) {
					ok = false;
					getBeanPotLogger().error("postConfiguration error", e); //$NON-NLS-1$
				}
		}
		if (!ok)
			throw new BeanPotException("post-configuration failed (see previous messages)"); //$NON-NLS-1$
	}

	/**
	 * Method to invoke {@link Configurable#preConfigure(Object)} on each bean
	 * in the container, with data passed in as the parameter.
	 * 
	 * @param data
	 *            arbitrary data for pre-configuration.
	 * @throws BeanPotException
	 * @see Configurable#preConfigure(Object)
	 */
	private void preConfigureBeans(final Object data) throws BeanPotException {
		getBeanPotLogger().debug("Pre-configure beans in container"); //$NON-NLS-1$
		boolean ok = true;
		for (final String key : getBeanKeys()) {
			final Object bean = getBean(key);
			// TODO consider doing this via polymorphism
			if (bean instanceof Configurable)
				try {
					((Configurable) bean).preConfigure(data);
				} catch (final BeanPotException e) {
					ok = false;
					getBeanPotLogger().error("preConfiguration error", e); //$NON-NLS-1$
				}
		}
		if (!ok)
			throw new BeanPotException("pre-configuration failed (see previous messages)"); //$NON-NLS-1$
	}

	/**
	 * @param parentConfig
	 * @param include
	 * @return
	 * @throws BeanPotException
	 */
	private Collection<HierarchicalConfiguration> processIncludeElement(final HierarchicalConfiguration parentConfig,
			final HierarchicalConfiguration include) throws BeanPotException {
		XMLConfiguration includeConfiguration = null;
		final Object value = include.getRootNode().getValue();
		getBeanPotLogger().info("Processing include file: " + value); //$NON-NLS-1$
		if (value instanceof String) {
			final String filename = (String) value;
			try {
				includeConfiguration = BeanConfiguration.create(filename);
			} catch (final ConfigurationException e) {
				if (parentConfig instanceof XMLConfiguration) {
					final String configUri = ((XMLConfiguration) parentConfig).getDocument().getDocumentURI();
					try {
						includeConfiguration = BeanConfiguration.create(new URL(new URL(configUri), filename));
					} catch (final MalformedURLException e1) {
						throw new BeanPotException("logic error", e1); //$NON-NLS-1$
					} catch (final ConfigurationException e2) {
						throw new BeanPotException(
								"BeanPot.addBean(HierarchicalConfiguration): include problem caused exception for: " + filename, e2); //$NON-NLS-1$
					}
				} else
					throw new BeanPotException("BeanPot.addBean(HierarchicalConfiguration): include problem caused exception", e); //$NON-NLS-1$
			}
		}
		if (includeConfiguration != null)
			return addBeans(includeConfiguration);
		return null;
	}

	/**
	 * @param serializer
	 * @param beanMap
	 * @param key
	 * @throws IOException
	 * @throws BeanPotException
	 */
	private void restoreBean(final Serializer serializer, final BeanMap beanMap, final String key) throws IOException,
			BeanPotException {
		final int available = serializer.available();
		if (available > 0) {
			try {
				beanMap.addBean(key, serializer.readObject(), Kleenean.Maybe);
			} catch (final ClassNotFoundException e) {
				throw new BeanPotException("restoreBean(): exception restoring: " + key, e); //$NON-NLS-1$
			}
		} else
			throw new BeanPotException("restoreBean(): no data available for: " + key); //$NON-NLS-1$
	}

	/**
	 * Private instance method to set the configuration according to an XML file
	 * (or other sourve of {@link HierarchicalConfiguration}). This method is
	 * called by {@link #setConfiguration(HierarchicalConfiguration)}.
	 * 
	 * @param config
	 */
	private void setConfig(final HierarchicalConfiguration config) {
		this.config = config;
	}

	/**
	 * @param configurationSource
	 *            the configurationSource to set
	 * 
	 */
	private void setConfigurationSource(final String configurationSource) {
		this.configurationSource = configurationSource;
	}

	/**
	 * Set this bean pot's configuration flag as given.
	 * 
	 * @param configured
	 */
	private void setConfigured(final boolean configured) {
		this.configured = configured;
	}

	/**
	 * @param identifier
	 */
	private void setIdentifier(final String identifier) {
		this.identifier = identifier;
		getBeanPotLogger().info("BeanPot identified as: " + identifier); //$NON-NLS-1$
	}

	/**
	 * Set the initialized flag according to the value provided.
	 * 
	 * @param initialized
	 */
	private void setInitialized(final boolean initialized) {
		this.initialized = initialized;
	}

	/**
	 * @param key
	 * @param writer
	 */
	private void showBean(final String key, final PrintWriter writer) {
		try {
			writer.print("Bean: " + key); //$NON-NLS-1$
			final Object bean = getBean(key);
			writer.println(" (" + bean.getClass().getCanonicalName() + ")"); //$NON-NLS-1$ //$NON-NLS-2$
			BeanUtilities.showProperties(bean, writer, getBeanMap());
		} catch (final BeanPotException e) {
			writer.println("** showBean exception: " + e.getLocalizedMessage()); //$NON-NLS-1$
			Throwable cause = e.getCause();
			while (cause != null) {
				writer.println("**** caused by: " + cause.getLocalizedMessage()); //$NON-NLS-1$
				cause = cause.getCause();
			}
		}
		writer.flush();
	}

	/**
	 * @param hcfg
	 *            a hierarchical configuration, typically from getting a
	 *            configuration "at" a key from a configuration, typically
	 *            instantiated as {@link XMLConfiguration}
	 */
	private void showConfig(final HierarchicalConfiguration hcfg) {
		if (getBeanPotLogger().isDebugEnabled()) {
			final Iterator<String> keys = hcfg.getKeys();
			while (keys.hasNext()) {
				final String key = keys.next();
				getBeanPotLogger().debug(key + "=" + hcfg.getProperty(key)); //$NON-NLS-1$
			}
		}
	}

	private static final HierarchicalConfiguration NULL_CONFIGURATION = null;

	/**
	 * 
	 */
	private static final String PATTERN_INCLUDE = "include"; //$NON-NLS-1$

	private static Object preConfigData = null;

	private static Object postConfigData = null;

	private static final String S_ATTR_PREFIX = "[@prefix]"; //$NON-NLS-1$

	/**
	 * The current reference to an MBeanServer. The type of this object should
	 * be javax.management.MBeanServer.
	 */
	private static Object MBeanServer;

	/**
	 * _MBeanPot_
	 */
	private static final String AGENT_MBEAN_POT_ = "_MBeanPot_"; //$NON-NLS-1$

	// Private constant class fields

	/**
	 * 
	 */
	private final static Map<String, String> _agentClassDefaults = new HashMap<String, String>();

	/**
	 * MBeanPot
	 */
	private static final String S_MBEAN_POT = "MBeanPot"; //$NON-NLS-1$

	/**
	 * id
	 */
	private static final String S_IDENTIFIER = "id"; //$NON-NLS-1$

	/**
	 * String recognized by BeanPot.getBean() method to yield the writer, as
	 * opposed to a bean. Value: _Writer_
	 */
	static final String AGENT_WRITER_ = "_Writer_"; //$NON-NLS-1$

	/**
	 * beans.bean
	 */
	private static final String S_BEANS_BEAN = "beans.bean"; //$NON-NLS-1$

	// Private variable class fields

	/**
	 * :
	 */
	private static final String S_COLON = ": "; //$NON-NLS-1$

	/**
	 * BeanPot.setConfiguration(File)
	 */
	private static final String S_M_SET_CONFIGURATION_FILE = "BeanPot.setConfiguration(File)"; //$NON-NLS-1$

	/**
	 * writer
	 */
	private static final String S_WRITER = "writer"; //$NON-NLS-1$

	/**
	 * If so-configured, this holds the reference to the MBeanPot.
	 */
	private static MBeanPot MBeanPot = null;

	/**
	 * This is the default prefix used when naming an MBean. The name is made up
	 * as: MBeanPrefix :type= beanName
	 */
	private static String MBeanPrefix = "com.rubecula.beanpot.jmx"; //$NON-NLS-1$

	// Private constant instance fields

	/**
	 * The (singleton) instance of this class (null until getInstance is first
	 * called).
	 */
	private static BeanPot Instance;

	/**
	 * The pattern that defines the beans in the configuration file. Default
	 * value is {@link #S_BEANS_BEAN}.
	 */
	private static String Pattern = S_BEANS_BEAN;

	/**
	 * Very simple logger, shared by all classes in this package.
	 */
	static Log beanPotLogger;

	/**
	 * unnamed bean pot
	 */
	private String identifier = "unnamed bean pot"; //$NON-NLS-1$

	// Private variable instance fields

	private String configurationSource;

	/**
	 * Map of this container's beans.
	 */
	private final BeanMap _beanMap = new BeanMap();

	/**
	 * Flag to remember if we have been initialized yet.
	 */
	private boolean initialized = false;

	/**
	 * Flag to remember if we have been configured yet.
	 */
	private boolean configured = false;

	/**
	 * the configuration of this BeanPot
	 */
	private HierarchicalConfiguration config;

	private static boolean validate = false;

	/**
	 * This location corresponds to the main/resources directory.
	 */
	private static final String DTD_LOCATION = "../../../beanpot.dtd"; //$NON-NLS-1$

	/**
	 * This is the default URL for the DTD.
	 */
	private static URL DTD = BeanConfiguration.class.getResource(BeanPot.DTD_LOCATION);

	/**
	 * Default print writer.
	 */
	private static PrintWriter printWriter = new PrintWriter(System.out, true);

}
