/**
 * BeanPot Bean Container Project.
 * Copyright (C) 2009  Rubecula Software, LLC.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Module: BeanPotLogger.java
 * Created on Mar 12, 2009
 * @version $Revision: 1.1 $
 */

package com.rubecula.beanpot;

import java.io.PrintWriter;
import java.text.DateFormat;
import java.util.Date;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.logging.Log;

/**
 * Very simple logger which extends the Java util logger.
 * 
 * There are only two useful levels: WARNING and DEBUG. It's compatible with
 * Jakarta Commons Logging in that TRACE and INFO are equivalent to DEBUG, and
 * FATAL and ERROR are equivalent to WARNING.
 * 
 * It is able to format the output as a pattern. The standard pattern is
 * {@link #STANDARD_PATTERN}:
 * 
 * Basic variables are of the form %Name%. A construction of the form $(...) is
 * an optional construct. It should include exactly one variable. If that
 * variable expands to nothing, the entire construct will be left out.
 * Otherwise, the contents of the construct (ellipsis above) will be used with
 * the variable appropriately substituted.
 * 
 * These are the supported variables:
 * <dl>
 * <dt>Date</dt>
 * <dd>the current date and time</dd>
 * <dt>Exception</dt>
 * <dd>the localized exception message</dd>
 * <dt>Level</dt>
 * <dd>the severity level</dd>
 * <dt>Message</dt>
 * <dd>message passed into the {@link Logger#log(Level, String)} method</dd>
 * <dt>Name</dt>
 * <dd>the name of this logger</dd>
 * </dl>
 * 
 * Example: (the standard pattern):
 * <code>&(%Level%:)%Name%:%Message%&(-%Exception%)</code>
 * 
 * The sub-pattern <code>&(%Level%:)</code> expands to either <i>WARNING:</i> or
 * the empty string (if Level is other than Warning); The sub-pattern
 * <code>&(%Exception%:)</code> expands to either
 * <code>e.getLocalizedMessage():</code> or the empty string (if there was no
 * exception passed in to the log method); The sub-pattern <code>%Name%</code>
 * expands to <code>getName()</code>; The sub-pattern <code>%Message%</code>
 * expands to <code>message</code> where message is the string passed into the
 * log method; <br>
 * 
 * @author Robin Hillyard
 * 
 */
public class BeanPotLogger extends Logger implements Log {

	private static Pattern getRegexPattern() {
		final String regexLiteralString = regexGroupZeroOrMore(RGX_CH_LITERAL);
		final String regexGroupVariable = RGX_CH_VAR_DELIM + regexGroupZeroOrMore(RGX_CH_LETTER) + RGX_CH_VAR_DELIM;
		final String regexGroupVarLit = regexGroupOnceIfAny(regexGroup(regexGroupVariable)) + regexLiteralString;
		final String regexGroupMetaVariable = regexGroupOnceIfAny(regexGroup(RGX_CH_META_DELIM_O))
				+ regexGroupOnceIfAny(regexLiteralString) + regexGroup(regexGroupVarLit) + regexGroup(RGX_CH_META_DELIM_C);
		final Pattern regex = Pattern.compile(regexGroupZeroOrMore(regexGroupMetaVariable));
		return regex;
	}

	/**
	 * Get the replacement string for a matched variable pattern.
	 * 
	 * @param matcher
	 * @param replacement
	 * @return
	 */
	private static String getReplacement(final Matcher matcher, final String replacement) {
		return matcher.group(RGX_GRP_VAR_PREFIX) + (replacement != null ? replacement : S_MT) + matcher.group(RGX_GRP_VAR_SUFFIX);
	}

	private static String regexGroup(final String regex) {
		return RGX_GRP_STRT + regex + RGX_GRP_END;
	}

	private static String regexGroupOnceIfAny(final String regex) {
		return regexGroup(regex + RGX_QU_ONCE_IF_ANY);
	}

	private static String regexGroupZeroOrMore(final String regex) {
		return regexGroup(regex + RGX_QU_ZERO_MORE);
	}

	/**
	 * Secondary constructor which invokes
	 * {@link #BeanPotLogger(String, PrintWriter)} with null for the print
	 * writer (i.e. uses the default).
	 * {@link #BeanPotLogger(String, String, PrintWriter)} with the null
	 * pattern.
	 * 
	 * @param name
	 *            the name of this logger (usually from the class using it).
	 */
	public BeanPotLogger(final String name) {
		this(name, NULL_WRITER);
	}

	/**
	 * Secondary constructor which invokes
	 * {@link #BeanPotLogger(String, String, PrintWriter)} with the null
	 * pattern.
	 * 
	 * @param name
	 *            the name of this logger (usually from the class using it).
	 * @param writer
	 *            the PrintWriter that will actually output the log messages.
	 */
	public BeanPotLogger(final String name, final PrintWriter writer) {
		this(name, NULL_PATTERN, writer);
	}

	/**
	 * Secondary constructor which invokes the primary constructor
	 * {@link #BeanPotLogger(String, PrintWriter, String, String)} with null for
	 * the pattern.
	 * 
	 * @param name
	 *            the name of this logger (usually from the class using it).
	 * @param writer
	 *            the PrintWriter that will actually output the log messages.
	 * @param resourceBundleName
	 *            the name of the resource bundle passed into the underlying
	 *            constructor (may be null).
	 */
	public BeanPotLogger(final String name, final PrintWriter writer, final String resourceBundleName) {
		this(name, writer, NULL_PATTERN, resourceBundleName);
	}

	/**
	 * Primary constructor which invokes the super-constructor with the
	 * parameters <code>name</code> and <code>resourceBundleName</code>.
	 * 
	 * @param name
	 *            the name of this logger (usually from the class using it).
	 * @param writer
	 *            the PrintWriter that will actually output the log messages
	 *            (may be null in which case it defaults to an instance of
	 *            {@link ConsoleWriter}.
	 * @param pattern
	 *            the pattern to use for outputting log entries (may be null in
	 *            which case it defaults to {@link #STANDARD_PATTERN}.
	 * @param resourceBundleName
	 *            the name of the resource bundle passed into the underlying
	 *            constructor (may be null).
	 */
	public BeanPotLogger(final String name, final PrintWriter writer, final String pattern, final String resourceBundleName) {
		super(name, resourceBundleName);
		if (writer != null)
			setWriter(writer);
		this._pattern = pattern != null ? pattern : STANDARD_PATTERN;
	}

	/**
	 * Secondary constructor which invokes
	 * {@link #BeanPotLogger(String, String, PrintWriter)} with null for the
	 * print writer.
	 * 
	 * @param name
	 *            the name of this logger (usually from the class using it).
	 * @param pattern
	 *            the pattern for the logged entries (may null in which case we
	 *            use the default).
	 */
	public BeanPotLogger(final String name, final String pattern) {
		this(name, pattern, NULL_WRITER);
	}

	/**
	 * Secondary constructor which invokes
	 * {@link #BeanPotLogger(String, PrintWriter, String, String)} with the null
	 * resource bundle name.
	 * 
	 * @param name
	 *            the name of this logger (usually from the class using it).
	 * @param writer
	 *            the PrintWriter that will actually output the log messages.
	 * @param pattern
	 *            the pattern for the logged entries (may null in which case we
	 *            use the default).
	 */
	public BeanPotLogger(final String name, final String pattern, final PrintWriter writer) {
		this(name, writer, pattern, NULL_RESOURCE_BUNDLE);
	}

	/**
	 * invoke {@link #log(Level, String)} with Level={@link #DEBUG} and
	 * message.toString()
	 * 
	 * @see org.apache.commons.logging.Log#debug(java.lang.Object)
	 */
	public void debug(final Object message) {
		log(DEBUG, message.toString());
	}

	/**
	 * invoke {@link #log(Level, String)} with Level={@link #DEBUG} and
	 * message.toString()
	 * 
	 * @see org.apache.commons.logging.Log#debug(java.lang.Object,
	 *      java.lang.Throwable)
	 */
	public void debug(final Object message, final Throwable t) {
		log(DEBUG, message.toString(), t);
	}

	/**
	 * invoke {@link #warn(Object)}
	 * 
	 * @see org.apache.commons.logging.Log#error(java.lang.Object)
	 */
	public void error(final Object message) {
		warn(message);
	}

	/**
	 * invoke {@link #warn(Object, Throwable)}
	 * 
	 * @see org.apache.commons.logging.Log#error(java.lang.Object,
	 *      java.lang.Throwable)
	 */
	public void error(final Object message, final Throwable t) {
		warn(message, t);
	}

	/**
	 * invoke {@link #debug(Object)}
	 * 
	 * @see org.apache.commons.logging.Log#fatal(java.lang.Object)
	 */
	public void fatal(final Object message) {
		warn(message);
	}

	/**
	 * invoke {@link #warn(Object, Throwable)}
	 * 
	 * @see org.apache.commons.logging.Log#fatal(java.lang.Object,
	 *      java.lang.Throwable)
	 */
	public void fatal(final Object message, final Throwable t) {
		warn(message, t);
	}

	/**
	 * invoke {@link #debug(Object)}
	 * 
	 * @see org.apache.commons.logging.Log#info(java.lang.Object)
	 */
	public void info(final Object message) {
		debug(message);
	}

	/**
	 * invoke {@link #debug(Object, Throwable)}
	 * 
	 * @see org.apache.commons.logging.Log#info(java.lang.Object,
	 *      java.lang.Throwable)
	 */
	public void info(final Object message, final Throwable t) {
		debug(message, t);
	}

	/**
	 * Method to discover if the debug mode is on.
	 * 
	 * @return the debug mode setting.
	 */
	public boolean isDebug() {
		return this.debug;
	}

	/**
	 * @return {@link #isDebug()}
	 * @see org.apache.commons.logging.Log#isDebugEnabled()
	 */
	public boolean isDebugEnabled() {
		return isDebug();
	}

	/**
	 * @return true
	 * @see org.apache.commons.logging.Log#isErrorEnabled()
	 */
	public boolean isErrorEnabled() {
		return true;
	}

	/**
	 * @return true
	 * @see org.apache.commons.logging.Log#isFatalEnabled()
	 */
	public boolean isFatalEnabled() {
		return true;
	}

	/**
	 * @return {@link #isDebugEnabled()}
	 * @see org.apache.commons.logging.Log#isInfoEnabled()
	 */
	public boolean isInfoEnabled() {
		return isDebugEnabled();
	}

	/**
	 * @return {@link #isDebugEnabled()}
	 * @see org.apache.commons.logging.Log#isTraceEnabled()
	 */
	public boolean isTraceEnabled() {
		return isDebugEnabled();
	}

	/**
	 * @return true
	 * @see org.apache.commons.logging.Log#isWarnEnabled()
	 */
	public boolean isWarnEnabled() {
		return true;
	}

	/**
	 * log a message if the level is {@link #WARNING} or if we are in debug
	 * mode.
	 * 
	 * @param level
	 *            the level (either {@link #WARNING} or {@link #DEBUG}).
	 * @param msg
	 *            the message to be logged
	 * @see java.util.logging.Logger#log(java.util.logging.Level,
	 *      java.lang.String)
	 */
	@Override
	public void log(final Level level, final String msg) {
		log(level, msg, NULL_THROWABLE);
	}

	/**
	 * log a message (inc. exception) if the level is {link #WARNING} or if we
	 * are in debug mode.
	 * 
	 * @param level
	 *            the level (either {@link #WARNING} or {@link #DEBUG}).
	 * @param msg
	 *            the message to be logged
	 * @param thrown
	 *            the exception to be logged (may be null)
	 * @see java.util.logging.Logger#log(java.util.logging.Level,
	 *      java.lang.String, java.lang.Throwable)
	 */
	@Override
	public void log(final Level level, final String msg, final Throwable thrown) {
		if (isDebug() || level == WARNING) {
			getWriter().println(prepareMessage(level, msg, thrown));
		}
	}

	/**
	 * log a message if we are in debug mode.
	 * 
	 * @param msg
	 */
	public void log(final String msg) {
		log(DEBUG, msg);
	}

	/**
	 * invoke {@link #log(Level, String, Throwable)} with level =
	 * {@link #WARNING}
	 * 
	 * @param msg
	 * @param thrown
	 */
	public void log(final String msg, final Throwable thrown) {
		log(WARNING, msg, thrown);
	}

	/**
	 * Set debug mode according to the value passed in.
	 * 
	 * @param debug
	 */
	public void setDebug(final boolean debug) {
		this.debug = debug;
	}

	/**
	 * Set a new {@link PrintWriter} for this logger.
	 * 
	 * @param writer
	 */
	public void setWriter(final PrintWriter writer) {
		this.writer = writer;
	}

	/**
	 * invoke {@link #debug(Object)}
	 * 
	 * @see org.apache.commons.logging.Log#trace(java.lang.Object)
	 */
	public void trace(final Object message) {
		debug(message);
	}

	/**
	 * invoke {@link #debug(Object, Throwable)}
	 * 
	 * @see org.apache.commons.logging.Log#trace(java.lang.Object,
	 *      java.lang.Throwable)
	 */
	public void trace(final Object message, final Throwable t) {
		debug(message, t);
	}

	/**
	 * invoke {@link #log(Level, String)} with Level={@link #WARNING} and
	 * message.toString()
	 * 
	 * @see org.apache.commons.logging.Log#warn(java.lang.Object)
	 */
	public void warn(final Object message) {
		log(WARNING, message.toString());
	}

	/**
	 * invoke {@link #log(Level, String, Throwable)} with Level={@link #WARNING}
	 * and message.toString()
	 * 
	 * @see org.apache.commons.logging.Log#warn(java.lang.Object,
	 *      java.lang.Throwable)
	 */
	public void warn(final Object message, final Throwable t) {
		log(WARNING, message.toString(), t);
	}

	/**
	 * Method to get the writer for this logger.
	 * 
	 * @return the current {@link PrintWriter} for this logger.
	 */
	PrintWriter getWriter() {
		return this.writer;
	}

	/**
	 * Run the method {@link Pattern#matcher(CharSequence)} on the
	 * {@link #_regexPattern} value, with parameter <code>pattern</code>; then
	 * call {@link #findAndSubstituteVariables(String, String, String, Matcher)}
	 * with the parameters passed in here.
	 * 
	 * @param msg
	 * @param levelString
	 * @param exceptionString
	 * @return the new string based on <code>pattern</code> but with the
	 *         variables substituted appropriately.
	 * 
	 *         XXX note that this has package scope only so that it can be
	 *         unit-tested
	 */
	String parseAndSubstituteVariables(final String msg, final String levelString, final String exceptionString) {
		final Matcher matcher = _regexPattern.matcher(getPattern());
		return findAndSubstituteVariables(msg, levelString, exceptionString, matcher);
	}

	/**
	 * Protected method which is called by the log() methods of this logger.
	 * 
	 * @param level
	 *            the severity level
	 * @param msg
	 *            the context-specific message to be logged
	 * @param thrown
	 *            an exception that necessitated the log (may be null)
	 * @return a string which will be sent to the writer.
	 */
	protected String prepareMessage(final Level level, final String msg, final Throwable thrown) {
		final String levelString = level != null ? (level == WARNING ? (level.toString() + S_MT) : null) : null;
		final String exceptionString = thrown != null ? "exception" + COLON + thrown.getLocalizedMessage() : null; //$NON-NLS-1$
		final StringBuilder result = new StringBuilder(parseAndSubstituteVariables(msg, levelString, exceptionString));
		if (thrown != null && thrown.getCause() != null)
			result.append("\n    " + prepareMessage(null, "caused by", thrown.getCause())); //$NON-NLS-1$ //$NON-NLS-2$
		return result.toString();
	}

	/**
	 * Given the matcher, find all of the appropriate groups and do
	 * substitutions as necessary.
	 * 
	 * @param msg
	 * @param levelString
	 * @param exceptionString
	 * @param matcher
	 * @return
	 */
	private String findAndSubstituteVariables(final String msg, final String levelString, final String exceptionString,
			final Matcher matcher) {
		// TODO note that group 5 is unnecessary (same as group 4)
		final StringBuilder result = new StringBuilder(S_MT);
		while (matcher.find()) {
			// showMatch(matcher);
			final String variableName = matcher.group(RGX_GRP_VARIABLE_NAME);
			if (variableName != null) {
				final String variableSubstitution = lookup(variableName, msg, levelString, exceptionString);
				if (matcher.group(RGX_GRP_META_PATTERN) != null) {
					if (variableSubstitution != null) {
						result.append(getReplacement(matcher, variableSubstitution));
					}
					// else do nothing
				} else
					result.append(getReplacement(matcher, variableSubstitution));
			} else
				result.append(matcher.group(RGX_GRP_PREFIX)); // literal
		}

		return result.toString();
	}

	/**
	 * @return the pattern
	 */
	private String getPattern() {
		return this._pattern;
	}

	/**
	 * Method to look up the value for a variable.
	 * 
	 * @param variable
	 * @param message
	 * @param level
	 * @param exception
	 * @return the string that corresponds to the variable.
	 */
	private String lookup(final String variable, final String message, final String level, final String exception) {
		if (variable.equalsIgnoreCase("Message")) //$NON-NLS-1$
			return message;
		if (variable.equalsIgnoreCase("Name")) //$NON-NLS-1$
			return getName();
		if (variable.equalsIgnoreCase("Level")) //$NON-NLS-1$
			return level;
		if (variable.equalsIgnoreCase("Exception")) //$NON-NLS-1$
			return exception;
		if (variable.equalsIgnoreCase("Date")) //$NON-NLS-1$
			return DateFormat.getDateTimeInstance().format(new Date());
		throw new RuntimeException("unrecognized variable: " + variable); //$NON-NLS-1$

	}

	/**
	 * Only used for debugging.
	 * 
	 * @param matcher
	 */
	@SuppressWarnings("unused")
	private void showMatch(final Matcher matcher) {
		System.out.println("found=\"" + matcher.group(0) + S_DBLE_QUOT); //$NON-NLS-1$
		final int groupCount = matcher.groupCount();
		for (int i = 0; i < groupCount; i++) {
			final String group = matcher.group(i + 1);
			System.out.println("group " + (i + 1) + S_EQLS + //$NON-NLS-1$
					S_DBLE_QUOT + group + S_DBLE_QUOT);
		}
	}

	/**
	 * 
	 */
	private static final Throwable NULL_THROWABLE = null;

	/**
	 * 
	 */
	private static final PrintWriter NULL_WRITER = null;

	private static final String NULL_PATTERN = null;

	private static final String NULL_RESOURCE_BUNDLE = null;

	/**
	 * 
	 */
	private static final String S_EQLS = "="; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String S_DBLE_QUOT = "\""; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final int RGX_GRP_VAR_SUFFIX = 10;

	/**
	 * 
	 */
	private static final int RGX_GRP_VAR_PREFIX = 4;

	/**
	 * 
	 */
	private static final int RGX_GRP_PREFIX = 1;

	/**
	 * 
	 */
	private static final int RGX_GRP_META_PATTERN = 3;

	/**
	 * 
	 */
	private static final int RGX_GRP_VARIABLE_NAME = 9;

	/**
	 * 
	 */
	private static final String RGX_META_ALT_C = "]"; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String RGX_META_ALT_O = "["; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String RGX_META_NOT = "^"; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String RGX_CH_VAR_DELIM = "%"; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String RGX_CH_META_DELIM_O = "\\&\\("; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String RGX_CH_META_DELIM_C = "\\)"; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String RGX_QU_ONCE_IF_ANY = "?"; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String RGX_QU_ZERO_MORE = "*"; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String RGX_GRP_END = ")"; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String RGX_GRP_STRT = "("; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String RGX_CH_LETTER = RGX_META_ALT_O + "A-Za-z" + RGX_META_ALT_C; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String RGX_CH_LITERAL = RGX_META_ALT_O + RGX_META_NOT + RGX_CH_META_DELIM_O + RGX_CH_META_DELIM_C
			+ RGX_CH_VAR_DELIM + RGX_META_ALT_C;

	/**
	 * 
	 */
	static final String S_MT = ""; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String STANDARD_PATTERN = "&(%Level%: )%Name%: %Message%&( -- %Exception%)"; //$NON-NLS-1$

	/**
	 * :
	 */
	protected static final String COLON = ": "; //$NON-NLS-1$

	private final String _pattern;

	/**
	 * {@link Level#WARNING}
	 */
	public static final Level WARNING = Level.WARNING;

	/**
	 * {@link Level#INFO}
	 */
	public static final Level DEBUG = Level.INFO;

	private PrintWriter writer = new ConsoleWriter();

	private boolean debug;

	private static final Pattern _regexPattern = getRegexPattern();

	/**
	 * 
	 */
	public static final String SPACE = " "; //$NON-NLS-1$

}
