/**
 * BeanPot BeanUtilities Container Project.
 * Copyright (C) 2009  Rubecula Software, LLC.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Module: BeanUtilities.java
 * Created on Jul 2, 2009
 * @version $Revision: 1.1 $
 */

package com.rubecula.beanpot;

import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.io.PrintWriter;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConversionException;
import org.apache.commons.configuration.HierarchicalConfiguration;
import org.apache.commons.configuration.SubnodeConfiguration;
import org.apache.commons.configuration.XMLConfiguration;
import org.apache.commons.configuration.tree.ConfigurationNode;
import org.apache.commons.logging.Log;

/**
 * This is basically a class for off-loading bean utilities from the BeanMap
 * class. Everything here could be in the BeanMap class (where it was) but then
 * that class becomes huge and it's hard to see what's really happening.
 * 
 * @author Robin Hillyard
 * 
 */
public class BeanUtilities {

	/**
	 * Construct (using appropriate constructor) an instance of the class
	 * defined by clazz and return it.
	 * 
	 * @param clazz
	 * @param classLoader
	 *            XXX
	 * @param identifier
	 *            string identifying the component which the instantiated object
	 *            will become
	 * @param args
	 *            the constructor args (null if looking for a singleton)
	 * @param acceptableException
	 *            XXX
	 * @return the newly instantiated object
	 * @throws BeanPotException
	 * 
	 */
	static Object constructBean(final String clazz, final ClassLoader classLoader, final String identifier, final Object[] args,
			final Class<? extends Throwable> acceptableException) throws BeanPotException {
		final Class<?> beanClass = getClassForString(clazz, classLoader);
		if (args == null)
			return getSingletonBean(beanClass, identifier);
		return constructBean(identifier, args, beanClass, acceptableException);
	}

	/**
	 * Get a class defined by <code>clazz</code>, but prefixed, if necessary, by
	 * <i>java.lang</i>.
	 * 
	 * @param clazz
	 * @param classLoader
	 *            XXX
	 * @return the class corresponding to clazz
	 * @throws BeanPotException
	 */
	static Class<?> getClassForString(final String clazz, final ClassLoader classLoader) throws BeanPotException {
		try {
			if (clazz.endsWith(S_BRACKETS)) {
				final String qualifiedName = getQualifiedClassName(clazz.substring(0, clazz.length() - 2));
				return Class.forName(S_ARRAY_CLASS_PREFIX + qualifiedName + S_ARRAY_CLASS_SUFFIX);
			}

			return classLoader.loadClass(getQualifiedClassName(clazz));
			// return Class.forName(getQualifiedClassName(clazz));
		} catch (final ClassNotFoundException e) {
			throw new BeanPotException("unable to interpret " + clazz + " as class name"); //$NON-NLS-1$ //$NON-NLS-2$
		}
	}

	/**
	 * @param bean
	 *            the bean on which we want to operate.
	 * @param name
	 *            the name of the property we are interested in.
	 * @return the property descriptor corresponding to <code>name</code>.
	 * @throws IntrospectionException
	 */
	static PropertyDescriptor getPropertyDescriptor(final Object bean, final String name) throws IntrospectionException {
		final PropertyDescriptor[] propertyDescriptors = Introspector.getBeanInfo(bean.getClass()).getPropertyDescriptors();
		for (int i = 0; i < propertyDescriptors.length; i++) {
			if (name.equalsIgnoreCase(propertyDescriptors[i].getName()))
				return propertyDescriptors[i];
		}
		return null;
	}

	/**
	 * @param beanMap
	 *            the bean map
	 * @param config
	 *            the definition of this bean in the configuration file
	 * @param key
	 *            the identifier of the bean which is to be made a member of
	 *            some other bean (must be non-null).
	 * @throws BeanPotException
	 */
	@SuppressWarnings({})
	static void makeBeanAMember(final BeanMap beanMap, final HierarchicalConfiguration config, final String key)
			throws BeanPotException {
		final String member = config.getString(""); //$NON-NLS-1$
		if (member != null) {
			final Object parentBean = beanMap.getBeanRef(member);
			if (parentBean != null && !(parentBean instanceof String)) {
				final Object bean = beanMap.getBean(key);
				if (bean != null) {
					if (addBeanAsMember(config, key, parentBean, bean, "makeBeanAMember(): bean " + key, beanMap)) //$NON-NLS-1$
						beanMap.getLogger().info("added bean with id " + key + " as a member of bean referenced by " + member); //$NON-NLS-1$ //$NON-NLS-2$
				} else
					throw new BeanPotException("makeBeanAsMember: bean key not found: " + key); //$NON-NLS-1$
			} else
				throw new BeanPotException("configuration for bean includes member with bad bean reference: " + member); //$NON-NLS-1$
		}
	}

	/**
	 * Method to allow external applications to dynamically set a bean's
	 * property.
	 * 
	 * XXX there is some overlap with other utilities.
	 * 
	 * @param bean
	 *            the bean on which to operate.
	 * @param value
	 *            the new value for the property.
	 * @param propName
	 *            the property name.
	 * @param beanId
	 *            the identifier of the bean (only used for error messages).
	 * 
	 * @throws BeanPotException
	 */
	static void setBeanProperty(final Object bean, final Object value, final String propName, final String beanId)
			throws BeanPotException {
		final String beanPropertyName = beanId + "." + propName; //$NON-NLS-1$
		try {
			final PropertyDescriptor propertyDescriptor = getPropertyDescriptor(bean, propName);
			if (propertyDescriptor != null) {
				final Method writeMethod = propertyDescriptor.getWriteMethod();
				if (writeMethod != null) {
					final Class<?> requiredClass = writeMethod.getParameterTypes()[0];
					try {
						Object parameter = value;
						if (!requiredClass.isAssignableFrom(value.getClass()))
							if (value instanceof String)
								parameter = BeanMap.getValueFromString(requiredClass, (String) value);
						writeMethod.invoke(bean, parameter);
					} catch (final IllegalArgumentException e) {
						throw new BeanPotException(
								"Incorrect argument for bean property " + beanPropertyName + " which expects " + //$NON-NLS-1$ //$NON-NLS-2$
										requiredClass.getName() + ": " + value.getClass().getName(), e); //$NON-NLS-1$
					}
				} else
					throw new BeanPotException("bean property is not writable: " + beanPropertyName); //$NON-NLS-1$
			} else
				throw new BeanPotException("bean property not found: " + beanPropertyName); //$NON-NLS-1$
		} catch (final IntrospectionException e) {
			throw new BeanPotException("Introspection exception on bean " + beanId, e); //$NON-NLS-1$
		} catch (final IllegalAccessException e) {
			throw new BeanPotException("Illegal access exception on bean property " + beanPropertyName, e); //$NON-NLS-1$
		} catch (final InvocationTargetException e) {
			throw new BeanPotException("Invocation target exception on bean property " + beanPropertyName, e); //$NON-NLS-1$
		}
	}

	/**
	 * @param bean
	 * @param writer
	 *            XXX
	 * @param beanMap
	 *            XXX
	 * @throws BeanPotException
	 */
	static void showProperties(final Object bean, final PrintWriter writer, final Map<String, Object> beanMap)
			throws BeanPotException {
		try {
			final PropertyDescriptor[] propertyDescriptors = Introspector.getBeanInfo(bean.getClass()).getPropertyDescriptors();
			for (int i = 0; i < propertyDescriptors.length; i++) {
				final PropertyDescriptor propertyDescriptor = propertyDescriptors[i];
				try {
					showProperty(bean, writer, beanMap, propertyDescriptor);
				} catch (final BeanPotException e) {
					if (propertyDescriptor.getWriteMethod() != null)
						throw e;
					writer.println(PREFIX_PROPERTY + propertyDescriptor.getName() + " ignored (not true property)"); //$NON-NLS-1$
				}
			}
		} catch (final IntrospectionException e) {
			throw new BeanPotException("BeanPot.showProperties: introspection exception", e); //$NON-NLS-1$
		}

	}

	/**
	 * @param config
	 * @param id
	 *            the bean's id (key)
	 * @param parentBean
	 * @param memberBean
	 * @param errorPrefix
	 *            a string which will be prefixed to any error messages.
	 * @param beanMap
	 *            XXX
	 * @return true if successful
	 * @throws BeanPotException
	 */
	private static boolean addBeanAsMember(final HierarchicalConfiguration config, final String id, final Object parentBean,
			final Object memberBean, final String errorPrefix, final BeanMap beanMap) throws BeanPotException {
		final String membership = getStringAttributeValue(config, S_MEMBERSHIP, errorPrefix, false);
		final String key = getStringAttributeValue(config, S_KEY, errorPrefix, true);
		final String value = getStringAttributeValue(config, S_VALUE, errorPrefix, true);

		final boolean isList = key == null && value == null;
		final boolean beanIsKey = isList ? false : key == null;
		final String prefix = isList ? S_ADD_METHOD_PREFIX : S_PUT_METHOD_PREFIX;
		final String methodName = prefix + membership.substring(0, 1).toUpperCase() + membership.substring(1);
		final ClassLoader classLoader = beanMap.getClassLoader();
		final Class<? extends Object> otherClass = getOtherClass(config, key, value, classLoader);
		final Method method = getAddMethod(methodName, !isList, parentBean.getClass(), memberBean.getClass(), otherClass,
				beanIsKey);
		if (method != null) {

			final Object otherObject = isList ? null : getOtherObject(otherClass, beanIsKey ? value : key, beanMap);
			return invokeAddMemberMethod(parentBean, memberBean, method, otherObject, "makeBeanAsMember(): bean " + id, //$NON-NLS-1$
					!isList, beanIsKey);
		}

		throw new BeanPotException(errorPrefix
				+ " cannot be made a " + membership + "member of " + parentBean + " (no appropriate method: " + methodName + ")"); //$NON-NLS-1$//$NON-NLS-2$//$NON-NLS-3$ //$NON-NLS-4$
	}

	/**
	 * Return true <b>if</b> desiredType is assignable from actualType or if
	 * desiredType is a primitive and actualType is the corresponding primitive
	 * where primitive may be one of int, double, float, long, short, char,
	 * byte.
	 * 
	 * @param desiredType
	 * @param actualType
	 * @return
	 */
	private static boolean checkCompatibleTypes(final Class<?> desiredType, final Class<?> actualType) {
		if (desiredType.isAssignableFrom(actualType))
			return true;
		if (desiredType == Integer.TYPE && actualType == Integer.class)
			return true;
		if (desiredType == Boolean.TYPE && actualType == Boolean.class)
			return true;
		if (desiredType == Double.TYPE && actualType == Double.class)
			return true;
		if (desiredType == Float.TYPE && actualType == Float.class)
			return true;
		if (desiredType == Long.TYPE && actualType == Long.class)
			return true;
		if (desiredType == Short.TYPE && actualType == Short.class)
			return true;
		if (desiredType == Character.TYPE && actualType == Character.class)
			return true;
		if (desiredType == Byte.TYPE && actualType == Byte.class)
			return true;
		return false;
	}

	/**
	 * Method to check that the <code>constructor</code> matches the
	 * <code>classes</code> provided. For each parameter type of
	 * <code>constructor</code>, we check that it is assignable from the
	 * corresponding (by position) class of <code>classes</code>.
	 * 
	 * @param constructor
	 *            the constructor to test
	 * @param classes
	 *            the classes of the parameters
	 * @return true if the constructor matches the classes
	 */
	private static boolean checkConstructorMatch(final Constructor<?> constructor, final Class<?>[] classes) {
		boolean result = true;
		for (int j = 0; result && j < classes.length; j++) {
			result = checkCompatibleTypes(constructor.getParameterTypes()[j], classes[j]);
		}
		return result;
	}

	/**
	 * Method to create a string corresponding to an array of classes.
	 * 
	 * @param classes
	 * @return the string in the form <code>(class,class,...class)</code>
	 */
	private static String classesToString(final Class<?>[] classes) {
		final StringBuilder result = new StringBuilder("("); //$NON-NLS-1$
		for (int i = 0; i < classes.length; i++) {
			result.append(classes[i].getName());
			if (i < classes.length - 1)
				result.append(","); //$NON-NLS-1$
		}
		result.append(")"); //$NON-NLS-1$
		return result.toString();
	}

	/**
	 * @param identifier
	 * @param args
	 * @param beanClass
	 * @param acceptableException
	 *            XXX
	 * @return
	 * @throws BeanPotException
	 */
	private static Object constructBean(final String identifier, final Object[] args, final Class<?> beanClass,
			final Class<? extends Throwable> acceptableException) throws BeanPotException {
		final Class<?>[] classes = new Class<?>[args.length];
		for (int i = 0; i < args.length; i++)
			classes[i] = args[i].getClass();
		try {
			final Constructor<?> constructor = getConstructor(identifier, beanClass, classes);
			return constructor.newInstance(args);
		} catch (final InstantiationException e) {
			throw new BeanPotException("instantiation expection for " + identifier, e); //$NON-NLS-1$
		} catch (final IllegalAccessException e) {
			throw new BeanPotException("illegal access exception for " + identifier, e); //$NON-NLS-1$
		} catch (final InvocationTargetException e) {
			final Throwable targetException = e.getTargetException();
			if (acceptableException != null && acceptableException.isAssignableFrom(targetException.getClass())) {
				System.err
						.println("ignored exception: " + targetException.getClass() + ": " + targetException.getLocalizedMessage()); //$NON-NLS-1$ //$NON-NLS-2$
				return null;
			}
			throw new BeanPotException("constructor threw exception for " + identifier, targetException); //$NON-NLS-1$
		}
	}

	/**
	 * @param requiredClass
	 * @param array
	 * @param listArray
	 */
	private static void fillPrimitiveDoubleArrayFromNumbers(final Class<? extends Number> requiredClass, final double[] array,
			final Object[] listArray) {
		for (int i = 0; i < array.length; i++) {
			array[i] = requiredClass.cast(listArray[i]).doubleValue();
		}
	}

	/**
	 * @param requiredType
	 * @param array
	 * @param stringArray
	 * @throws BeanPotException
	 */
	private static void fillPrimitiveDoubleArrayFromStrings(final Class<? extends Number> requiredType, final double[] array,
			final Object[] stringArray) throws BeanPotException {
		for (int i = 0; i < array.length; i++) {
			array[i] = requiredType.cast(BeanMap.getValueFromString(requiredType, (String) stringArray[i])).doubleValue();
		}
	}

	/**
	 * @param requiredClass
	 * @param array
	 * @param listArray
	 */
	private static void fillPrimitiveIntArrayFromNumbers(final Class<? extends Number> requiredClass, final int[] array,
			final Object[] listArray) {
		for (int i = 0; i < array.length; i++) {
			array[i] = requiredClass.cast(listArray[i]).intValue();
		}
	}

	/**
	 * @param requiredType
	 * @param array
	 * @param stringArray
	 * @throws BeanPotException
	 */
	private static void fillPrimitiveIntArrayFromStrings(final Class<? extends Number> requiredType, final int[] array,
			final Object[] stringArray) throws BeanPotException {
		for (int i = 0; i < array.length; i++) {
			array[i] = requiredType.cast(BeanMap.getValueFromString(requiredType, (String) stringArray[i])).intValue();
		}
	}

	/**
	 * @param beanMap
	 * @param value
	 * @return
	 */
	private static String findKey(final Map<String, Object> beanMap, final Object value) {
		for (final Entry<String, Object> entry : beanMap.entrySet())
			if (value.equals(entry.getValue()))
				return entry.getKey();
		return null;
	}

	/**
	 * @param memberBean
	 * @param key
	 * @param isMap
	 * @param isKey
	 * @return
	 */
	private static Object[] getAddMemberParameters(final Object memberBean, final Object key, final boolean isMap,
			final boolean isKey) {
		final List<Object> parameters = new LinkedList<Object>();
		if (isMap)
			parameters.add(key);
		parameters.add(isKey ? 0 : parameters.size(), memberBean);
		return parameters.toArray();
	}

	/**
	 * Find a method in the <code>parentClass</code> whose name is "addXxxx" or
	 * "putXxxx" where "Xxxx" is derived from "xxxx" being the
	 * <code>membership</code> passed in. The method must take one parameter
	 * which is assignable from <code>memberClass</code> and an optional (first)
	 * parameter which is a String (they key).
	 * 
	 * @param methodName
	 *            in the form of either addXxxx or putXxxx
	 * @param isMap
	 *            XXX
	 * @param parentClass
	 *            the class of the parent-to-be.
	 * @param valueClass
	 *            the value class for the thing being added.
	 * @param keyClass
	 *            the key class (null if not a map)
	 * @param isKey
	 *            true if the bean to be added is the key to a map (as opposed
	 *            to the value) (ignored if isMap is false).
	 * @return the method which will be used to add member to parent.
	 * @throws BeanPotException
	 */
	private static Method getAddMethod(final String methodName, final boolean isMap, final Class<? extends Object> parentClass,
			final Class<? extends Object> valueClass, final Class<? extends Object> keyClass, final boolean isKey)
			throws BeanPotException {
		Method result = null;
		final Method[] methods = parentClass.getMethods();
		for (int i = 0; i < methods.length; i++) {
			final Method method = methods[i];
			if (method.getName().equals(methodName)) {
				final Class<?>[] parameterTypes = method.getParameterTypes();
				final int params = isMap ? 2 : 1;
				if (parameterTypes.length == params) {
					// TODO consider using checkCompatibleTypes() here and below
					boolean ok = checkCompatibleTypes(parameterTypes[isKey ? 0 : params - 1], valueClass);
					// boolean ok = ((Class<?>) parameterTypes[isKey ? 0 :
					// params - 1])
					// .isAssignableFrom(valueClass);
					if (isMap)
						ok &= checkCompatibleTypes(parameterTypes[isKey ? params - 1 : 0], keyClass);
					// ok &= ((Class<?>) parameterTypes[isKey ? params - 1 :
					// 0]).isAssignableFrom(keyClass);
					if (ok)
						result = method;
				}
			}
		}
		if (result != null)
			return result;
		throw new BeanPotException(
				"no method named " + methodName + " found in " + parentClass + " which takes parameter(s) of type " + (isMap ? keyClass.toString() : "") + ", " + valueClass + (isKey ? " or reversed" : "")); //$NON-NLS-1$//$NON-NLS-2$//$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$
	}

	/**
	 * @param config
	 * @param classLoader
	 *            XXX
	 * @return
	 * @throws BeanPotException
	 */
	@SuppressWarnings("unchecked")
	private static Class<?> getClassFromAttribute(final SubnodeConfiguration config, final ClassLoader classLoader)
			throws BeanPotException {
		Class<?> clazz = null;
		final Iterator<String> keys = config.getKeys();
		while (keys.hasNext()) {
			final String key = keys.next();
			if (key.equalsIgnoreCase(S_ATTR_CLASS)) {
				// if the class is specified explicitly then use that of course
				clazz = getClassForString(config.getString(key), classLoader);
			}
		}
		return clazz;
	}

	/**
	 * Method to get a suitable constructor with the appropriate matching
	 * classes for the constructor arguments. First we try calling the obvious
	 * {@link Class#getConstructor(Class...)} method but that doesn't always
	 * give us what we want. Therefore, if a {@link NoSuchMethodException} is
	 * thrown, we take charge and look for a suitable constructor ourselves.
	 * 
	 * @param identifier
	 * @param beanClass
	 * @param classes
	 * @return
	 * @throws BeanPotException
	 * @see BeanUtilities#checkConstructorMatch(Constructor, Class[])
	 */
	private static Constructor<?> getConstructor(final String identifier, final Class<?> beanClass, final Class<?>[] classes)
			throws BeanPotException {
		try {
			return beanClass.getConstructor(classes);
		} catch (final NoSuchMethodException e) {
			Constructor<?> constructor = null;
			final Constructor<?>[] constructors = beanClass.getConstructors();
			for (int i = 0; constructor == null && i < constructors.length; i++) {
				final Constructor<?> cnstcr = constructors[i];
				if (cnstcr.getParameterTypes().length == classes.length) {
					if (checkConstructorMatch(cnstcr, classes))
						constructor = cnstcr;
				}
			}

			if (constructor != null)
				return constructor;
			throw new BeanPotException("getConstructor()" + //$NON-NLS-1$
					S_COLON
					+ "No appropriate constructor for " + identifier + ": " + beanClass.getName() + classesToString(classes), e); //$NON-NLS-1$ //$NON-NLS-2$ 

		}
	}

	/**
	 * @param listArray
	 * @param listType
	 * @param componentType
	 * @return
	 * @throws BeanPotException
	 */
	private static Object getObjectArray(final Object[] listArray, final Class<? extends Object> listType,
			final Class<?> componentType) throws BeanPotException {
		final Object[] array = (Object[]) Array.newInstance(componentType, listArray.length);

		for (int i = 0; i < array.length; i++) {
			// TODO consider using checkCompatibleTypes()
			if (componentType.isAssignableFrom(listType))
				array[i] = listArray[i];
			else if (listType == String.class)
				array[i] = BeanMap.getValueFromString(componentType, (String) listArray[i]);
			else
				throw new BeanPotException("incompatible list type: " + listType + " for array property type: " //$NON-NLS-1$ //$NON-NLS-2$
						+ componentType);
		}
		return array;
	}

	/**
	 * @param config
	 * @param key
	 * @param value
	 * @param classLoader
	 * @return
	 * @throws BeanPotException
	 */
	@SuppressWarnings("unchecked")
	private static Class<?> getOtherClass(final HierarchicalConfiguration config, final String key, final String value,
			final ClassLoader classLoader) throws BeanPotException {
		Class<?> valueClass = Object.class;
		Class<?> keyClass = String.class;
		final Iterator<String> propertyAttrs = config.getKeys();
		while (propertyAttrs.hasNext()) {
			final String object = propertyAttrs.next();
			if (value != null && object.equalsIgnoreCase(S_ATTR_VALUE_CLASS))
				valueClass = getClassForString(config.getString(object), classLoader);
			if (key != null && object.equalsIgnoreCase(S_ATTR_KEY_CLASS))
				keyClass = getClassForString(config.getString(object), classLoader);
		}
		return key != null ? keyClass : value != null ? valueClass : null;
	}

	/**
	 * @param otherClass
	 * @param keyString
	 * @param beanMap
	 *            XXX
	 * @return
	 * @throws BeanPotException
	 */
	private static Object getOtherObject(final Class<? extends Object> otherClass, final String keyString, final BeanMap beanMap)
			throws BeanPotException {
		final Object beanRef = beanMap.getBeanRef(keyString);
		if (beanRef != null && checkCompatibleTypes(otherClass, beanRef.getClass()))
			return beanRef;
		return BeanMap.getValueFromString(otherClass, keyString);
	}

	/**
	 * Get the appropriate class for the parameter described by
	 * <code>config</code>.
	 * 
	 * @param config
	 *            the configuration describing this parameter
	 * @param classLoader
	 *            XXX
	 * @return either the explicitly defined class (using the "class" attribute)
	 *         or the default class which is {@link Integer} if the value can
	 *         thus be represented, otherwise it is {@link String}.
	 * @throws BeanPotException
	 */
	private static Class<?> getParameterClass(final SubnodeConfiguration config, final ClassLoader classLoader)
			throws BeanPotException {
		// Set default clazz to be String
		Class<?> clazz = String.class;
		try {
			// Try to see if class appears to be an integer
			config.getInt(S_VALUE);
			clazz = Integer.class;
		} catch (final ConversionException e1) {
			// do nothing, we assume for now that it's a String
		}
		final Class<?> classFromAttribute = getClassFromAttribute(config, classLoader);
		if (classFromAttribute != null)
			clazz = classFromAttribute;
		return clazz;
	}

	/**
	 * Create and return a primitive array, such as int[] or double[], from a
	 * list of objects, which must be either {@link Number}s or {@link String}s.
	 * 
	 * XXX This method (and its own private methods) could probably be
	 * refactored a bit.
	 * 
	 * @param listArray
	 * @param listType
	 * @param componentType
	 * @throws BeanPotException
	 */
	private static Object getPrimitiveArray(final Object[] listArray, final Class<? extends Object> listType,
			final Class<?> componentType) throws BeanPotException {
		if (componentType == Integer.TYPE) {
			final int[] array = (int[]) Array.newInstance(componentType, listArray.length);
			if (listType == Integer.class)
				fillPrimitiveIntArrayFromNumbers(Integer.class, array, listArray);
			else if (listType == String.class)
				fillPrimitiveIntArrayFromStrings(Integer.class, array, listArray);
			else
				throw new BeanPotException("incompatible list type: " + listType + " for primitive array property type: " //$NON-NLS-1$ //$NON-NLS-2$
						+ componentType);
			return array;
		} else if (componentType == Double.TYPE) {
			final double[] array = (double[]) Array.newInstance(componentType, listArray.length);
			if (listType == Double.class)
				fillPrimitiveDoubleArrayFromNumbers(Double.class, array, listArray);
			else if (listType == String.class)
				fillPrimitiveDoubleArrayFromStrings(Double.class, array, listArray);
			else
				throw new BeanPotException("incompatible list type: " + listType + " for primitive array property type: " //$NON-NLS-1$ //$NON-NLS-2$
						+ componentType);
			return array;
		}
		throw new BeanPotException("unsupported primitive array type (currently must be int or double): " + componentType); //$NON-NLS-1$
	}

	/**
	 * @param clazz
	 * @return
	 */
	private static String getQualifiedClassName(final String clazz) {
		return (clazz.contains(".") ? clazz : S_JAVA_LANG + clazz); //$NON-NLS-1$
	}

	/**
	 * @param defaultClass
	 *            this is the class that will be returned if the className is
	 *            null
	 * @param specifiedClass
	 *            this class (provided as a String) will be returned provided
	 *            that it is not null and it is compatible with required class,
	 *            in particular, requiredClass.isAssignableFrom(specifiedClass)
	 * @param property
	 *            only used by the exception constructor (if an error occurs).
	 * @param classLoader
	 *            XXX
	 * @return
	 * @throws BeanPotException
	 */
	private static Class<?> getRequiredClass(final Class<?> defaultClass, final String specifiedClass, final String property,
			final ClassLoader classLoader) throws BeanPotException {
		Class<?> requiredClass = defaultClass;
		if (specifiedClass != null) {
			final Class<?> clazzSpec = getClassForString(specifiedClass, classLoader);
			// TODO consider using checkCompatibleTypes()
			if (requiredClass.isAssignableFrom(clazzSpec))
				requiredClass = clazzSpec;
			else
				throw new BeanPotException(M_SET_BEAN_PROPERTY + S_COLON
						+ "property " + property + " cannot be assigned a value of the specified class: " + specifiedClass); //$NON-NLS-1$//$NON-NLS-2$
		}
		return requiredClass;
	}

	/**
	 * @param beanClass
	 * @param identifier
	 *            the bean identifier (used if exception is thrown)
	 * @return the bean
	 * @throws BeanPotException
	 */
	private static Object getSingletonBean(final Class<?> beanClass, final String identifier) throws BeanPotException {
		try {
			final Method method = beanClass.getMethod("getInstance", new Class[0]); //$NON-NLS-1$
			return method.invoke(NULL_OBJECT, new Object[0]);
		} catch (final IllegalAccessException e) {
			throw new BeanPotException("illegal access exception for " + identifier, e); //$NON-NLS-1$
		} catch (final InvocationTargetException e) {
			throw new BeanPotException("getInstance() threw exception for " + identifier, e.getTargetException()); //$NON-NLS-1$
		} catch (final SecurityException e) {
			throw new BeanPotException("illegal access exception for " + identifier, e); //$NON-NLS-1$
		} catch (final NoSuchMethodException e) {
			throw new BeanPotException("illegal access exception for " + identifier, e); //$NON-NLS-1$
		}
	}

	/**
	 * @param config
	 * @param name
	 * @param errorPrefix
	 *            a String to use in any exceptions.
	 * @param nullOk
	 *            XXX
	 * @return the value of the attribute called name, as a String.
	 * @throws BeanPotException
	 *             either missing, redundant or non-String attribute value.
	 */
	@SuppressWarnings("unchecked")
	private static String getStringAttributeValue(final HierarchicalConfiguration config, final String name,
			final String errorPrefix, final boolean nullOk) throws BeanPotException {
		String result = null;
		for (final ConfigurationNode attribute : (List<ConfigurationNode>) config.getRootNode().getAttributes(name)) {
			final Object value = attribute.getValue();
			if (value instanceof String) {
				if (result == null)
					result = (String) value;
				else
					throw new BeanPotException(errorPrefix + ": has redundant attribute named: " + name); //$NON-NLS-1$
			} else {
				throw new BeanPotException(errorPrefix + ": has attribute " + name + //$NON-NLS-1$
						"which is not a String: " + value); //$NON-NLS-1$
			}
		}
		if (nullOk || result != null)
			return result;
		throw new BeanPotException(errorPrefix + ": has no attribute named: " + name); //$NON-NLS-1$

	}

	/**
	 * @param parentBean
	 * @param memberBean
	 * @param method
	 * @param key
	 *            the is the key value that will be used for a put method
	 *            invocation (and the member bean will go as the value)
	 * @param errorPrefix
	 *            an error prefix for exceptions
	 * @param isMap
	 *            XXX
	 * @param isKey
	 *            XXX
	 * @return true if successful
	 * @throws BeanPotException
	 */
	private static boolean invokeAddMemberMethod(final Object parentBean, final Object memberBean, final Method method,
			final Object key, final String errorPrefix, final boolean isMap, final boolean isKey) throws BeanPotException {
		try {
			final Object result = method.invoke(parentBean, getAddMemberParameters(memberBean, key, isMap, isKey));
			if (isMap)
				return result == null;
			else if (result instanceof Boolean)
				return ((Boolean) result).booleanValue();
			else
				return result != null;
		} catch (final IllegalAccessException e) {
			throw new BeanPotException(errorPrefix + ": illegal access exception with method " + method.getName(), e); //$NON-NLS-1$
		} catch (final InvocationTargetException e) {
			throw new BeanPotException(
					errorPrefix + ": method " + method.getName() + " threw an exception", e.getTargetException()); //$NON-NLS-1$//$NON-NLS-2$
		}
	}

	/**
	 * @param bean
	 * @param writer
	 * @param beanMap
	 * @param propertyDescriptor
	 * @throws BeanPotException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	private static void showProperty(final Object bean, final PrintWriter writer, final Map<String, Object> beanMap,
			final PropertyDescriptor propertyDescriptor) throws BeanPotException {
		final String beanName = BeanUtilities.findKey(beanMap, bean);
		final String name = propertyDescriptor.getDisplayName();
		final Method method = propertyDescriptor.getReadMethod();
		if (method != null) {
			try {
				final Object value = method.invoke(bean, new Object[0]);
				if (value == null)
					writer.println(PREFIX_PROPERTY + name + "=<null>"); //$NON-NLS-1$
				else if (beanMap.containsValue(value))
					writer.println(PREFIX_PROPERTY + name + "=!" + BeanUtilities.findKey(beanMap, value)); //$NON-NLS-1$
				else if (!"class".equals(name)) //$NON-NLS-1$
					writer.println(PREFIX_PROPERTY + name + "=" + value.toString()); //$NON-NLS-1$
			} catch (final IllegalAccessException e) {
				throw new BeanPotException(
						"BeanPot.showProperties: illegal access exception getting property: " + name + " of bean " + beanName, e); //$NON-NLS-1$ //$NON-NLS-2$
			} catch (final InvocationTargetException e) {
				throw new BeanPotException(
						"BeanPot.showProperties: target exception getting property: " + name + " of bean " + beanName, e.getTargetException()); //$NON-NLS-1$ //$NON-NLS-2$
			}
		}
	}

	/**
	 * Package-scope constructor. This should be invoked from the BeanMap class
	 * with "this" as the parameter.
	 * 
	 * @param beanMap
	 *            the bean map that constructed us.
	 * 
	 */
	BeanUtilities(final BeanMap beanMap) {
		super();
		this._beanMap = beanMap;
	}

	/**
	 * Instantiate a bean of the appropriate class and add it using
	 * <code>id</code> as the key
	 * 
	 * @param key
	 *            the key which will be used to reference the bean, i.e. the
	 *            identifier of the bean
	 * @param clazz
	 *            the class of the bean
	 * @param parameters
	 *            the constructor parameters (or null if the class is a
	 *            singleton class)
	 * @param registerAsMBean
	 *            if true or maybe, we try to register the bean as an MBean
	 * @param required
	 *            XXX
	 * @param acceptableException
	 *            XXX
	 * @return the newly added bean
	 * 
	 * @throws BeanPotException
	 * 
	 * @see BeanMap#addBean(String, Object, Kleenean) also
	 *      {@link #constructBean(String, ClassLoader, String, Object[], Class)}
	 */
	Object addBean(final String key, final String clazz, final Object[] parameters, final Kleenean registerAsMBean,
			final boolean required, final Class<? extends Throwable> acceptableException) throws BeanPotException {
		if (clazz != null) {
			final Object bean = constructBean(clazz, getBeanMap().getClassLoader(), S_BEAN + S_COLON + key, parameters,
					acceptableException);
			if (bean != null) {
				final Object oldBean = getBeanMap().addBean(key, bean, registerAsMBean);
				if (oldBean != null)
					getLogger().warn("Duplicate beans for " + key); //$NON-NLS-1$
				else
					getLogger().info("added bean " + key + " to container"); //$NON-NLS-1$ //$NON-NLS-2$
			} else if (required)
				throw new BeanPotException("required bean not created: " + key); //$NON-NLS-1$
			else
				getLogger().warn("omitted non-required bean: " + key + " (see trace for exception)"); //$NON-NLS-1$ //$NON-NLS-2$
			return bean;
		}
		throw new BeanPotException(M_ADD_BEAN + S_COLON + "no class defined for bean: " + key); //$NON-NLS-1$
	}

	/**
	 * Configure the bean identified by <code>id</code>, according to the
	 * <code>config</code> given. By "configure", we mean set the various bean
	 * properties by calling setX() methods.
	 * 
	 * @param id
	 *            bean identifier
	 * @param config
	 *            a hierarchical configuration, typically from getting a
	 *            configuration "at" a key from a configuration, typically
	 *            instantiated as {@link XMLConfiguration}
	 * 
	 * @throws IntrospectionException
	 * @throws BeanPotException
	 * 
	 */
	void configureBean(final String id, final HierarchicalConfiguration config) throws IntrospectionException, BeanPotException {
		final Object bean = getBeanMap().getBean(id, false);
		if (bean != null) {
			final PropertyDescriptor[] propertyDescriptors = Introspector.getBeanInfo(bean.getClass()).getPropertyDescriptors();
			try {
				configureBeanAttributes(bean, config, propertyDescriptors);
				configureBeanProperties(bean, config, propertyDescriptors, id);
			} catch (final BeanPotException e) {
				throw new BeanPotException("Problem configuring bean: " + id, e); //$NON-NLS-1$
			}
		} else
			getLogger().warn("configureBean()" + S_COLON + "bean undefined: " + id); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Configure the properties of the given <code>bean</code>. A configurable
	 * bean property is defined as one where there is a matching pair of get/set
	 * methods (although the set method may not be visible) and where the type
	 * of the return/parameter is derivable using methods such as
	 * {@link Configuration#getBoolean(String)} from the string value.
	 * 
	 * @param bean
	 * @param config
	 * @param propertyDescriptors
	 * @param id
	 *            the bean identifier
	 * 
	 * @throws BeanPotException
	 * 
	 */
	@SuppressWarnings("unchecked")
	void configureBeanProperties(final Object bean, final HierarchicalConfiguration config,
			final PropertyDescriptor[] propertyDescriptors, final String id) throws BeanPotException {
		final List<SubnodeConfiguration> propertyConfigurations = config.configurationsAt(S_PROPERTY);
		for (final SubnodeConfiguration propertyConfig : propertyConfigurations) {
			String propertyName = null;
			String className = null;
			final Iterator<String> keys = propertyConfig.getKeys();
			while (keys.hasNext()) {
				final String key = keys.next();
				final String value = propertyConfig.getString(key);
				if (key.equalsIgnoreCase(S_ATTR_NAME))
					propertyName = value;
				else if (key.equalsIgnoreCase(S_ATTR_CLASS))
					className = value;
			}
			if (propertyName != null) {
				if (!setBeanProperty(bean, S_VALUE, propertyConfig, propertyDescriptors, propertyName, className))
					getLogger().warn("property \"" + propertyName + "\" not found for bean: " + id); //$NON-NLS-1$//$NON-NLS-2$
			} else
				throw new BeanPotException("property has no name"); //$NON-NLS-1$
		}
	}

	/**
	 * @param key
	 * @param clazz
	 * @param config
	 * @return a primitive array of the appropriate type
	 * @throws BeanPotException
	 */
	Object getArrayParameter(final String key, final Class<?> clazz, final Configuration config) throws BeanPotException {
		final Collection<? extends Object> list = getListParameter(S_LIST + key, config);
		if (list.size() == 0)
			throw new BeanPotException("array list is empty"); //$NON-NLS-1$
		final Object[] listArray = list.toArray(new Object[0]);
		final Class<? extends Object> listType = listArray[0].getClass();
		final Class<?> componentType = clazz.getComponentType();
		if (listType != String.class && !checkCompatibleTypes(componentType, listType))
			throw new BeanPotException("incompatible list type: " + listType + " for array property type: " //$NON-NLS-1$ //$NON-NLS-2$
					+ componentType);

		if (componentType.isPrimitive())
			return getPrimitiveArray(listArray, listType, componentType);

		return getObjectArray(listArray, listType, componentType);
	}

	/**
	 * Get the constructor arguments for a new bean
	 * 
	 * @param config
	 * 
	 * @return an array of constructor arguments.
	 * @throws BeanPotException
	 * 
	 * @see BeanMap#addBeanRefToArgs(List, SubnodeConfiguration)
	 * 
	 * @see BeanUtilities#addParameterToArgs(List, SubnodeConfiguration)
	 */
	@SuppressWarnings("unchecked")
	Object[] getConstructorArgs(final HierarchicalConfiguration config) throws BeanPotException {
		final List<Object> args = new ArrayList<Object>();
		final List<SubnodeConfiguration> argConfigurations = config.configurationsAt(S_ARG);
		for (final SubnodeConfiguration argConfig : argConfigurations) {
			if (argConfig.configurationsAt(S_REF).size() == 1) {
				final SubnodeConfiguration beanRef = argConfig.configurationAt(S_REF);
				getBeanMap().addBeanRefToArgs(args, beanRef);
			} else if (argConfig.configurationsAt(S_VALUE).size() == 1) {
				addParameterToArgs(args, argConfig);
			} else if (argConfig.configurationsAt(S_LIST).size() == 1) {
				final SubnodeConfiguration subnodeConfiguration = argConfig.configurationAt(S_LIST);
				addListToArgs(args, subnodeConfiguration, getClassFromAttribute(argConfig, getBeanMap().getClassLoader()));
			}
		}
		return args.toArray();
	}

	/**
	 * This method will yield a List object from the <code>config</code>
	 * provided. The result can match any parameter whose type extends
	 * {@link Collection}&lt;?&gt;. For syntax, see
	 * {@link BeanMap#configureBean(HierarchicalConfiguration)}
	 * 
	 * @param key
	 * @param config
	 * 
	 * @return an object which is an {@link ArrayList}&lt;?&gt; where the
	 *         generic type ? is determined by the <i>valueclass</i> attribute
	 *         and defaults to Object.
	 * @throws BeanPotException
	 * 
	 */
	@SuppressWarnings("unchecked")
	Collection<? extends Object> getListParameter(final String key, final Configuration config) throws BeanPotException {
		final Iterator<String> propertyAttrs = config.getKeys();
		Class<?> valueClass = Object.class;
		while (propertyAttrs.hasNext()) {
			final String object = propertyAttrs.next();
			if (object.equalsIgnoreCase(S_ATTR_VALUE_CLASS)) {
				valueClass = getClassForString(config.getString(object), getBeanMap().getClassLoader());
			}
		}

		// TODO at this point, we should look for possible definition
		// of valueClass in the list entity itself (like we do for maps)
		final List<String> list = config.getList(key);
		final List<Object> result = new ArrayList<Object>();
		for (final String string : list) {
			// TODO consider using checkCompatibleTypes() instead of
			// isAssignableFrom in several places
			if (Integer.class.isAssignableFrom(valueClass))
				result.add(Integer.valueOf(string));
			else if (Long.class.isAssignableFrom(valueClass))
				result.add(Long.valueOf(string));
			else if (Double.class.isAssignableFrom(valueClass))
				result.add(Double.valueOf(string));
			else {
				final Object object = getStringAsBean(string);
				if (object != null) {
					if (valueClass.isAssignableFrom(object.getClass()))
						result.add(object);
					else {
						final Object valueFromString = BeanMap.getValueFromString(valueClass, string);
						if (valueFromString != null)
							result.add(valueFromString);
						else
							throw new BeanPotException("getListParameter(): null value returned from string: " + string); //$NON-NLS-1$
					}
				} else
					throw new BeanPotException("getListParameter(): null object defined by " + string); //$NON-NLS-1$
			}
		}
		return result;
	}

	/**
	 * This method will yield a List object from the <code>config</code>
	 * provided. The result can match any parameter whose type extends
	 * {@link Map}&lt;?,?&gt;.
	 * 
	 * For syntax, see {@link BeanMap#configureBean(HierarchicalConfiguration)}
	 * 
	 * All generic types which are supported by
	 * {@link BeanMap#getPropertyValue(String, Class, Configuration)} are
	 * supported (unlike the List case which only supports certain types).
	 * 
	 * TODO support use of getValueFromString for map values (currently, it
	 * doesn't work properly).
	 * 
	 * @param config
	 * 
	 * @return an object which is an {@link HashMap}&lt;?,?&gt; where the
	 *         generic types ? are determined by the <i>keyclass</i> and
	 *         <i>valueclass</i> attributes and default to String, Object,
	 *         respectively.
	 * @throws BeanPotException
	 * 
	 *             XXX probably could do some refactoring on this method.
	 */
	@SuppressWarnings("unchecked")
	Object getMapParameter(final HierarchicalConfiguration config) throws BeanPotException {
		final ClassLoader classLoader = getBeanMap().getClassLoader();
		Class<?> valueClass = Object.class;
		Class<?> keyClass = String.class;

		// First, we look through the attributes of the property entity for
		// valueClass and keyClass designations.
		final Iterator<String> propertyAttrs = config.getKeys();
		while (propertyAttrs.hasNext()) {
			final String object = propertyAttrs.next();
			if (object.equalsIgnoreCase(S_ATTR_VALUE_CLASS)) {
				valueClass = getClassForString(config.getString(object), classLoader);
			} else if (object.equalsIgnoreCase(S_ATTR_KEY_CLASS)) {
				keyClass = getClassForString(config.getString(object), classLoader);
			}
		}

		// Next, we look through the attributes of the map entity for
		// valueClass and keyClass designations.
		final SubnodeConfiguration mapConfig = config.configurationAt(S_MAP);
		@SuppressWarnings("rawtypes")
		final Iterator mapKeys = mapConfig.getKeys();
		while (mapKeys.hasNext()) {
			final String mapAttr = (String) mapKeys.next();
			if (mapAttr.equalsIgnoreCase(S_ATTR_KEY_CLASS)) {
				keyClass = getClassForString(mapConfig.getString(mapAttr), classLoader);
			} else if (mapAttr.equalsIgnoreCase(S_ATTR_VALUE_CLASS)) {
				valueClass = getClassForString(mapConfig.getString(mapAttr), classLoader);
			}
		}

		// Finally, we try to get the key and value themselves.
		final List<SubnodeConfiguration> subnodes = config.configurationsAt(S_MAP + ".entry");//$NON-NLS-1$
		final Map<Object, Object> result = new HashMap<Object, Object>();
		for (final SubnodeConfiguration subnode : subnodes) {
			Object key = null;
			Object value = null;
			@SuppressWarnings("rawtypes")
			final Iterator keys = subnode.getKeys();
			while (keys.hasNext()) {
				final String attrKey = (String) keys.next();
				if (attrKey.equalsIgnoreCase(S_ATTR_KEY)) {
					key = getBeanMap().getPropertyValue(attrKey, keyClass, subnode);
				} else if (attrKey.equalsIgnoreCase(S_ATTR_VALUE)) {
					value = getBeanMap().getPropertyValue(attrKey, valueClass, subnode);
				}
			}

			// If the value is still null, it means it's defined using an
			// element or just text instead of an attribute.
			if (value == null)
				value = getBeanMap().getPropertyValue("", valueClass, subnode); //$NON-NLS-1$
			if (value == null)
				value = getBeanMap().getPropertyValue("value", valueClass, subnode); //$NON-NLS-1$
			if (value == null)
				value = getBeanMap().getPropertyValue(".", valueClass, subnode); //$NON-NLS-1$
			if (key != null && value != null)
				result.put(key, value);
			else
				throw new BeanPotException(
						"getMapParameter(): map entry key/value is not defined (key=" + key + ",value=" + value + ")"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}
		return result;
	}

	/**
	 * Interpret the given string as a bean reference, thus returning the bean,
	 * else return the string itself. The string is a bean reference if it
	 * begins with "!" but not "!!" which is interpreted as the character "!".
	 * 
	 * @param string
	 *            a literal string or a bean name prefixed by "!".
	 * 
	 * @return either the referenced bean or the literal string.
	 */
	Object getStringAsBean(final String string) {
		if (string.startsWith(S_SHRIEK)) {
			if (string.startsWith(S_SHRIEK + S_SHRIEK))
				return string.substring(1);
			return getBeanMap().getBean(string.substring(1));
		}
		return string;
	}

	/**
	 * Using the setter method for the given <code>bean</code> (from the
	 * <code>propertyDescriptors</code>), set the property with given name to
	 * the value defined by the <code>config</code> (using <code>key</code> as
	 * the key).
	 * 
	 * @param bean
	 *            the bean whose parameter is to be set.
	 * @param key
	 *            the key which points to the <code>config</code> to provide the
	 *            value.
	 * @param config
	 *            the {@link Configuration} defining the property
	 * @param propertyDescriptors
	 *            the property descriptors (from introspecting <code>bean</code>
	 *            )
	 * @param propertyName
	 *            the property name
	 * @param className
	 *            optional parameter which, if specified, is the name of the
	 *            actual class or property value being provided
	 * @return XXX
	 * @throws BeanPotException
	 * 
	 */
	boolean setBeanProperty(final Object bean, final String key, final Configuration config,
			final PropertyDescriptor[] propertyDescriptors, final String propertyName, final String className)
			throws BeanPotException {
		// XXX this is a bit inefficient!
		boolean found = false;
		// TODO use method #getPropertyDescriptor() to get the matching property
		// descriptor
		for (int i = 0; i < propertyDescriptors.length; i++) {
			final PropertyDescriptor propertyDescriptor = propertyDescriptors[i];
			if (propertyName.equalsIgnoreCase(propertyDescriptor.getName())) {
				setBeanProperty(bean, config, key, propertyDescriptor, className);
				found = true;
			}
		}
		return found;
	}

	/**
	 * Method to flesh out the list of constructor arguments (<code>args</code>)
	 * when a list is found.
	 * 
	 * If <code>parameterClass</code> is an array, then that the resulting args
	 * list will have just one element: an array of the required type, filled
	 * with the list members [be careful, a class cast exception could be thrown
	 * in this case];
	 * 
	 * If <code>parameterClass</code> is a list, then the resulting args list
	 * will have just one element: an ArrayList, filled with the list members;
	 * 
	 * Otherwise, the resulting args list will be made up of each of the list
	 * members separately.
	 * 
	 * @param args
	 * @param argConfig
	 * @param parameterClass
	 * @throws BeanPotException
	 */
	private void addListToArgs(final List<Object> args, final SubnodeConfiguration argConfig, final Class<?> parameterClass)
			throws BeanPotException {
		final Collection<? extends Object> listParameter = getListParameter(S_VALUE, argConfig);
		if (parameterClass != null && parameterClass.isArray()) {
			final Object[] array = listParameter.toArray((Object[]) java.lang.reflect.Array.newInstance(
					parameterClass.getComponentType(), 0));
			args.add(array);
		}
		// TODO consider using checkCompatibleTypes()
		else if (parameterClass != null && parameterClass.isAssignableFrom(List.class)) {
			final List<Object> list = new ArrayList<Object>();
			for (final Object object : listParameter) {
				list.add(object);
			}
			args.add(list);
		} else
			for (final Object object : listParameter) {
				args.add(object);
			}
	}

	/**
	 * Method to add a parameter (i.e. not a bean reference) to the list of
	 * constructor <code>args</code>. The parameter is described by the subnode
	 * configuration <code>argConfig</code>. There is a restriction: the class
	 * of a value must be an object, not a primitive. If the value appears to be
	 * an integer, then by default its class will be {@link Integer}. Else by
	 * default, its class will be {@link String}. However, the class can always
	 * be explicitly set by using the class attribute.
	 * 
	 * @param args
	 *            the list of constructor args, to which we will add the new
	 *            constructor-arg.
	 * @param argConfig
	 *            the configuration subnode for new constructor-arg.
	 * @throws BeanPotException
	 * 
	 * @see #getPropertyValue(String, Class, Configuration)
	 */
	private void addParameterToArgs(final List<Object> args, final SubnodeConfiguration argConfig) throws BeanPotException {
		args.add(getBeanMap().getPropertyValue(S_VALUE, getParameterClass(argConfig, getBeanMap().getClassLoader()), argConfig));
	}

	/**
	 * Method to configure the given <code>bean</code> according to the given
	 * <code>config</code>. If the attribute is neither the <b>id</b> attribute
	 * or the <b>class</b> attribute, and provided it conforms to the standard
	 * naming form: <code>[@attr]</code> then the corresponding property name
	 * will be passed into
	 * {@link #setBeanProperty(Object, String, Configuration, PropertyDescriptor[], String, String)}
	 * 
	 * TODO consider removing this method and disallowing this type of property
	 * setting.
	 * 
	 * @param bean
	 *            the bean whose properties are to be set
	 * @param config
	 *            a hierarchical configuration, typically from getting a
	 *            configuration "at" a key from a configuration, typically
	 *            instantiated as {@link XMLConfiguration}
	 * @param propertyDescriptors
	 *            the property descriptors for the bean (found from
	 *            introspection).
	 * @param key
	 *            the key which defines the property value in the configuration
	 * @param propertyDescriptor
	 *            the result of introspecting the bean for properties
	 * @throws BeanPotException
	 * 
	 */
	private void configureAttribute(final Object bean, final Configuration config,
			final PropertyDescriptor[] propertyDescriptors, final String key) throws BeanPotException {
		// TODO figure out why we don't need to filter on mbean but do on
		// singleton
		if (key.startsWith("[@") && key.endsWith("]") && !key.equalsIgnoreCase(S_ATTR_ID) && !key.equalsIgnoreCase(S_ATTR_TYPE) && !key.equalsIgnoreCase(S_ATTR_CLASS)) { //$NON-NLS-1$ //$NON-NLS-2$
			final String property = key.substring(2, key.length() - 1);
			setBeanProperty(bean, key, config, propertyDescriptors, property, NULL_CLASSNAME);
		}
	}

	/**
	 * Configure the bean according to the attributes. Currently, attributes are
	 * passed in as if they were properties.
	 * 
	 * TODO consider changing things so that the attributes are either ignored
	 * or have some different outcome.
	 * 
	 * @param bean
	 * @param config
	 * @param propertyDescriptors
	 * @throws BeanPotException
	 * 
	 * @see BeanUtilities#configureAttribute(Object, Configuration,
	 *      PropertyDescriptor[], String)
	 */
	@SuppressWarnings("unchecked")
	private void configureBeanAttributes(final Object bean, final HierarchicalConfiguration config,
			final PropertyDescriptor[] propertyDescriptors) throws BeanPotException {
		final Iterator<String> keys = config.getKeys();
		while (keys.hasNext()) {
			configureAttribute(bean, config, propertyDescriptors, keys.next());
		}
	}

	/**
	 * @return the beanMap
	 */
	private BeanMap getBeanMap() {
		return this._beanMap;
	}

	/**
	 * @return
	 */
	private Log getLogger() {
		return getBeanMap().getLogger();
	}

	/**
	 * Set a property in the given bean.
	 * 
	 * @param bean
	 *            the bean whose property is to be set
	 * @param config
	 *            a hierarchical configuration, typically from getting a
	 *            configuration "at" a key from a configuration, typically
	 *            instantiated as {@link XMLConfiguration}
	 * @param key
	 *            the key which defines the property value in the configuration
	 * @param writeMethod
	 *            the method which is used to set the property value
	 * @param propertyClass
	 *            the class of the property
	 * 
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 * @throws BeanPotException
	 * 
	 * @see #getPropertyValue(String, Class, Configuration)
	 * 
	 * @see Method#invoke(Object, Object...)
	 */
	private void setBeanProperty(final Object bean, final Configuration config, final String key, final Method writeMethod,
			final Class<?> propertyClass) throws IllegalAccessException, InvocationTargetException, BeanPotException {
		final Object object = getBeanMap().getPropertyValue(key, propertyClass, config);

		if (object != null) {
			writeMethod.invoke(bean, new Object[] { object });
		} else
			throw new BeanPotException(M_SET_BEAN_PROPERTY + S_COLON
					+ "bean reference is non-existent for property" + S_COLON + writeMethod.getName() //$NON-NLS-1$
					+ " for key" + S_COLON + key + " where value is" + S_COLON + config.getString(key)); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Set a property in the given <code>bean</code>.
	 * 
	 * @param bean
	 *            the bean whose property is to be set
	 * @param config
	 *            a {@link Configuration} defined by a configuration file.
	 * @param key
	 *            the key which defines the property value in the configuration
	 * @param propertyDescriptor
	 *            the result of introspecting the bean for properties
	 * @param className
	 *            optional parameter which, if specified, is the name of the
	 *            actual class or property value being provided
	 * @throws BeanPotException
	 * 
	 * @see {@link #setBeanProperty(Object, Configuration, String, String, Method, String)}
	 */
	private void setBeanProperty(final Object bean, final Configuration config, final String key,
			final PropertyDescriptor propertyDescriptor, final String className) throws BeanPotException {
		final Method writeMethod = propertyDescriptor.getWriteMethod();
		if (writeMethod != null) {
			setBeanProperty(bean, config, key, propertyDescriptor.getName(), writeMethod, className);
		} else
			throw new BeanPotException(M_SET_BEAN_PROPERTY + S_COLON + "property is read-only" + //$NON-NLS-1$
					S_COLON + propertyDescriptor.getName());
	}

	/**
	 * Set a property in the given <code>bean</code>.
	 * 
	 * @param bean
	 *            the bean whose property is to be set
	 * @param config
	 *            a hierarchical configuration, typically from getting a
	 *            configuration "at" a key from a configuration, typically
	 *            instantiated as {@link XMLConfiguration}
	 * @param key
	 *            the key which defines the property value in the configuration
	 * @param propertyName
	 *            the name of the property
	 * @param writeMethod
	 *            the method which is used to set the property value
	 * @param className
	 *            optional parameter which, if specified, is the name of the
	 *            actual class or property value being provided
	 * @throws BeanPotException
	 * 
	 */
	private void setBeanProperty(final Object bean, final Configuration config, final String key, final String propertyName,
			final Method writeMethod, final String className) throws BeanPotException {
		final Class<?>[] parameterTypes = writeMethod.getParameterTypes();
		if (parameterTypes.length == 1) {
			try {
				final Class<?> requiredClass = getRequiredClass(parameterTypes[0], className, propertyName, getBeanMap()
						.getClassLoader());
				setBeanProperty(bean, config, key, writeMethod, requiredClass);
			} catch (final IllegalAccessException e) {
				throw new BeanPotException(M_SET_BEAN_PROPERTY + S_COLON + "illegal access bean property: " //$NON-NLS-1$
						+ propertyName, e);
			} catch (final InvocationTargetException e) {
				throw new BeanPotException(M_SET_BEAN_PROPERTY + S_COLON + "illegal method invocation exception: " //$NON-NLS-1$
						+ propertyName, e.getTargetException());
			}
		} else
			throw new BeanPotException(M_SET_BEAN_PROPERTY + S_COLON + "multiple parameters for bean property: " //$NON-NLS-1$
					+ propertyName);
	}

	/**
	 * Used for invocations of class methods.
	 */
	private static final Object NULL_OBJECT = null;

	private static final String NULL_CLASSNAME = null;

	private static final String S_KEY = "key"; //$NON-NLS-1$

	private static final String S_PUT_METHOD_PREFIX = "put"; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String S_MEMBERSHIP = "membership"; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String S_ADD_METHOD_PREFIX = "add"; //$NON-NLS-1$

	/**
	 * ;
	 */
	private static final String S_ARRAY_CLASS_SUFFIX = ";"; //$NON-NLS-1$

	/**
	 * [L
	 */
	private static final String S_ARRAY_CLASS_PREFIX = "[L"; //$NON-NLS-1$

	/**
	 * []
	 */
	private static final String S_BRACKETS = "[]"; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String PREFIX_PROPERTY = "    "; //$NON-NLS-1$

	/**
	 * bean
	 */
	static final String S_BEAN = "bean"; //$NON-NLS-1$

	/**
	 * [@id]
	 */
	static final String S_ATTR_ID = "[@id]"; //$NON-NLS-1$

	/**
	 * !
	 */
	static final String S_SHRIEK = "!"; //$NON-NLS-1$

	/**
	 * map
	 */
	private static final String S_MAP = "map"; //$NON-NLS-1$

	/**
	 * Default package for finding classes: java.lang.
	 */
	private static final String S_JAVA_LANG = "java.lang."; //$NON-NLS-1$

	/**
	 * list.
	 */
	public static final String S_LIST = "list."; //$NON-NLS-1$

	/**
	 * [@valueclass]
	 */
	private static final String S_ATTR_VALUE_CLASS = "[@valueclass]"; //$NON-NLS-1$

	/**
	 * [@keyclass]
	 */
	private static final String S_ATTR_KEY_CLASS = "[@keyclass]"; //$NON-NLS-1$

	/**
	 * [@required]
	 */
	static final String S_ATTR_REQUIRED = "[@required]"; //$NON-NLS-1$

	/**
	 * [@acceptableException]
	 */
	static final String S_ATTR_ACCEPTABLE_EXCEPTION = "[@acceptableException]"; //$NON-NLS-1$

	/**
	 * [@key]
	 */
	static final String S_ATTR_KEY = "[@key]"; //$NON-NLS-1$

	/**
	 * [@value]
	 */
	private static final String S_ATTR_VALUE = "[@value]"; //$NON-NLS-1$

	/**
	 * The tag for the definition of the value.
	 */
	private static final String S_VALUE = "value"; //$NON-NLS-1$

	/**
	 * This is what an attribute's configuration key looks like: [@name]
	 */
	static final String S_ATTR_NAME = "[@name]"; //$NON-NLS-1$

	/**
	 * ref
	 */
	private static final String S_REF = "ref"; //$NON-NLS-1$

	/**
	 * property
	 */
	private static final String S_PROPERTY = "property"; //$NON-NLS-1$

	/**
	 * constructor-arg
	 */
	private static final String S_ARG = "constructor-arg"; //$NON-NLS-1$

	/**
	 * addBean()
	 */
	private static final String M_ADD_BEAN = "addBean()"; //$NON-NLS-1$

	/**
	 * setBeanProperty()
	 */
	private static final String M_SET_BEAN_PROPERTY = "setBeanProperty()"; //$NON-NLS-1$

	/**
	 * :
	 */
	static final String S_COLON = ": "; //$NON-NLS-1$

	/**
	 * [@class]
	 */
	static final String S_ATTR_CLASS = "[@class]"; //$NON-NLS-1$

	/**
	 * This is the bean map from which were were originally constructed. It's
	 * slightly ugly to do it this way, but the alternative is to have one
	 * gigantic {@link BeanMap} class with all of these utility methods in it.
	 */
	private final BeanMap _beanMap;

	/**
	 * [@type]
	 */
	static final String S_ATTR_TYPE = "[@type]"; //$NON-NLS-1$

}
