/**
 * BeanPot Bean Container Project.
 * Copyright (C) 2009  Rubecula Software, LLC.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Module: Serializer.java
 * Created on Jul 27, 2010
 * @version $Revision: 1.1 $
 */

package com.rubecula.beanpot;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;


/**
 * Interface to aid in (de)-serialization of objects to/from {@link File} or
 * other data source/sink. <br/>
 * The whole process of de/serialization is quite complex to describe. Here's a
 * reference to the Java 1.5 PDF description of the: <a
 * href="http://java.sun.com/j2se/1.5/pdf/serial-1.5.0.pdf">Java Object
 * Serialization Specification</a> <br/>
 * However, it is extremely simply to use. You really don't need to use the
 * {@link Serializer} interface and classes at all. However, they are
 * convenient. The most basic scheme of usage, for example to serialize a single
 * object (perhaps a bean), is as follows:
 * 
 * <pre>
 * final Serializer sin = SerializerFactory.createSerializer(&quot;tempFile&quot;);
 * sin.writeObject(x);
 * </pre>
 * 
 * where x is some object to be serialized into the file called "tempFile". The
 * SerializerFactory has several methods which will create an appropriate
 * serializer.</br> The typical usage of serialization in the BeanPot can be
 * seen by looking at the test method:
 * The following calls are
 * provided:
 * <ul>
 * <li>{@link BeanPot#serializeBeans(Serializer)} and
 * {@link BeanPot#restoreBeans(Serializer, String[])} (the two most general
 * methods);</li>
 * <li>{@link BeanPot#serializeBeans(File)} and its counterpart
 * {@link BeanPot#serializeBeans(String)};</li>
 * <li>{@link BeanPot#restoreBeans(File, String[], boolean)} and its counterpart
 * {@link BeanPot#restoreBeans(String, String[], boolean)}.</li>
 * </ul>
 * If you use the simplest methods (where you simply provide a filename to be
 * used), a sequence might look like this:
 * 
 * <pre>
 * final String[] beanKeys = BeanPot.serializeBeans(&quot;tempBeans&quot;);
 * <i>... other code, possibly closing and restarting the BeanPot ...</i>
 * BeanPot.restoreBeans(&quot;tempBeans&quot;, beanKeys, true);
 * </pre>
 * 
 * The array returned from the serializeBeans method is an array of bean keys
 * (the strings which allow the beans to be looked up in the container). This is
 * used to get the corresponding objects back from the file and replace them in
 * the container.
 * 
 * The {@link BeanPot} class itself is not {@link Serializable} so that if a
 * bean contains non-transient references to the container (that should not
 * happen) or other non-serializable beans, problems will arise.</br> An aspect
 * of the serialization specification which is not intuitively obvious is that
 * it knows how to serialize a graph, meaning that you don't have to worry about
 * recursive references. See, for example, the
 * {@link SerializerTest#testSerializerRecursive()} method.
 * 
 * @author Robin Hillyard
 * 
 */
public interface Serializer {

	/**
	 * @return the number of bytes available to read
	 * @throws IOException
	 * @see java.io.ObjectInputStream#available()
	 */
	public abstract int available() throws IOException;

	/**
	 * @throws IOException
	 * @see java.io.ObjectOutputStream#close()
	 */
	public abstract void close() throws IOException;

	/**
	 * @throws IOException
	 * @see java.io.OutputStream#flush()
	 */
	public abstract void flush() throws IOException;

	/**
	 * @return a de-serialized object
	 * @throws IOException
	 * @throws ClassNotFoundException
	 * @see java.io.ObjectInputStream#readObject()
	 */
	public abstract Object readObject() throws IOException, ClassNotFoundException;

	/**
	 * @param obj
	 * @throws IOException
	 * @see java.io.ObjectOutputStream#writeObject(java.lang.Object)
	 */
	public abstract void writeObject(final Object obj) throws IOException;

}
